# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.sqs`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging
import re

from tornado import concurrent
from tornado import gen
from tornado import ioloop
import boto.sqs.connection
import boto.sqs.queue
import mock

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.aws import settings as aws_settings
from kingpin.actors.utils import dry
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class QueueNotFound(exceptions.RecoverableActorFailure):

    """Raised by SQS Actor when a needed queue is not found."""


class QueueDeletionFailed(exceptions.RecoverableActorFailure):

    """Raised if Boto fails to delete an SQS queue.

    http://boto.readthedocs.org/en/latest/ref/
        sqs.html#boto.sqs.connection.SQSConnection.delete_queue
    """


class SQSBaseActor(base.AWSBaseActor):

    # This actor should not be instantiated, but unit testing requires that
    # it's all options are defined properly here.
    all_options = {
        'name': (str, REQUIRED, 'Queue name to do nothing with.'),
        'region': (str, REQUIRED, 'AWS region (or zone) name like us-west-2')
    }

    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()
    executor = EXECUTOR

    @gen.coroutine
    def _fetch_queues(self, pattern):
        """Searches SQS for all queues with a matching name pattern.

        Args:
            pattern: string - regex used in `re.match()`

        Returns:
            Array of matched queues, even if empty.
        """
        queues = yield self.api_call(self.sqs_conn.get_all_queues)
        match_queues = [q for q in queues if re.search(pattern, q.name)]
        raise gen.Return(match_queues)


class Create(SQSBaseActor):

    """Creates a new SQS queue with the specified name

    **Options**

    :name:
      (str) The name of the queue to create

    :region:
      (str) AWS region (or zone) string, like 'us-west-2'

    **Examples**

    .. code-block:: json

       { "actor": "aws.sqs.Create",
         "desc": "Create queue named async-tasks",
         "options": {
           "name": "async-tasks",
           "region": "us-east-1",
         }
       }

    **Dry Mode**

    Will not create any queue, or even contact SQS. Will create a mock.Mock
    object and exit with success.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name or pattern for SQS queues.'),
        'region': (str, REQUIRED, 'AWS region (or zone), such as us-west-2')
    }

    @gen.coroutine
    def _create_queue(self, name):
        """Create an SQS queue with the specified name.

        Returns either the real boto.sqs.queue.Queue object or the Mock object
        in dry run.

        Args:
            name: Queue name (string) to create.

        Returns:
            An SQS Queue Object
        """
        if not self._dry:
            self.log.info('Creating a new queue: %s' % name)
            new_queue = yield self.api_call(self.sqs_conn.create_queue, name)
        else:
            self.log.info('Would create a new queue: %s' % name)
            new_queue = mock.Mock(name=name)

        self.log.debug('Returning queue object: %s' % new_queue)
        raise gen.Return(new_queue)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        Raises:
            gen.Return()
        """
        q = yield self._create_queue(name=self.option('name'))

        if q.__class__ == boto.sqs.queue.Queue:
            self.log.info('Queue Created: %s' % q.url)
        elif self._dry:
            self.log.info('Fake Queue: %s' % q)
        else:
            raise exceptions.UnrecoverableActorFailure(
                'All hell broke loose: %s' % q)

        raise gen.Return()


class Delete(SQSBaseActor):

    """Deletes the SQS queues

    Note: **even if it`s not empty**

    **Options**

    :name:
      (str) The name of the queue to destroy

    :region:
      (str) AWS region (or zone) string, like 'us-west-2'

    :idempotent:
      (bool) Will not raise errors if no matching queues are found.
      (default: False)

    **Examples**

    .. code-block:: json

       { "actor": "aws.sqs.Delete",
         "desc": "Delete queue async-tasks",
         "options": {
           "name": "async-tasks",
           "region": "us-east-1"
         }
       }

    .. code-block:: json

       { "actor": "aws.sqs.Delete",
         "desc": "Delete queues with 1234 in the name",
         "options": {
           "name": "1234",
           "region": "us-east-1"
         }
       }

    **Dry Mode**

    Will find the specified queue, but will have a noop regarding its deletion.
    Dry mode will fail if no queues are found, and idempotent flag is set to
    False.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name or pattern for SQS queues.'),
        'region': (str, REQUIRED, 'AWS region (or zone), such as us-west-2'),
        'idempotent': (bool, False, 'Continue if queues are already deleted.')
    }

    @gen.coroutine
    @dry('Would delete the queue: {queue}')
    def _delete_queue(self, queue):
        """Delete the provided queue.

        Raises RecoverableActorFailure if fail to delete it.

        Returns:
          True if successful in deletion, or is Dry run.

        Raises:
          QueueDeletionFailed if queue deletion failed.
        """
        self.log.info('Deleting Queue: %s...' % queue.url)
        ok = yield self.api_call(self.sqs_conn.delete_queue, queue)

        # Raise an exception if the tasks failed
        if not ok:
            raise QueueDeletionFailed('Failed to delete "%s"' % queue.url)

        raise gen.Return(ok)

    @gen.coroutine
    @utils.retry(QueueNotFound, delay=aws_settings.SQS_RETRY_DELAY)
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        Raises:
            gen.Return()
            QueueNotFound()
        """
        pattern = self.option('name')
        matched_queues = yield self._fetch_queues(pattern=pattern)

        not_found_condition = (not matched_queues and
                               not self.option('idempotent'))

        if not_found_condition:
            raise QueueNotFound(
                'No queues with pattern "%s" found.' % pattern)

        self.log.info('Deleting SQS Queues: %s' % matched_queues)

        tasks = []
        for q in matched_queues:
            tasks.append(self._delete_queue(queue=q))
        yield tasks

        raise gen.Return()


class WaitUntilEmpty(SQSBaseActor):

    """Wait indefinitely until for SQS queues to become empty

    This actor will loop infinitely as long as the count of messages in at
    least one queue is greater than zero. SQS does not guarantee exact count,
    so this can return a stale value if the number of messages in the queue
    changes rapidly.


    **Options**

    :name:
      (str) The name or regex pattern of the queues to operate on

    :region:
      (str) AWS region (or zone) string, like 'us-west-2'

    :required:
      (bool) Fail if no matching queues are found.
      (default: False)

    **Examples**

    .. code-block:: json

       { "actor": "aws.sqs.WaitUntilEmpty",
         "desc": "Wait until release-0025a* queues are empty",
         "options": {
           "name": "release-0025a",
           "region": "us-east-1",
           "required": true
         }
       }

    **Dry Mode**

    This actor performs the finding of the queue, but will pretend that the
    count is 0 and return success. Will fail even in dry mode if `required`
    option is set to True and no queues with the name pattern are found.
    """

    all_options = {
        'name': (str, REQUIRED, 'Name or pattern for SQS queues.'),
        'region': (str, REQUIRED, 'AWS region (or zone), such as us-west-2'),
        'required': (bool, False, 'At least 1 queue must be found.')
    }

    @gen.coroutine
    def _wait(self, queue, sleep=3):
        """Sleeps until an SQS Queue has emptied out.

        Args:
            queue: AWS SQS Queue object
            sleep: Int of seconds to wait between checks

        Returns:
            True: When queue is empty.
        """

        count = 0
        while True:
            if not self._dry:
                self.log.debug('Counting %s' % queue.url)
                visible = yield self.api_call(queue.count)
                attr = 'ApproximateNumberOfMessagesNotVisible'
                invisible = yield self.api_call(queue.get_attributes, attr)
                invisible_int = int(invisible[attr])
                count = visible + invisible_int
            else:
                self.log.info('Pretending that count is 0 for %s' % queue.url)
                count = 0

            self.log.debug('Queue has %s messages in it.' % count)
            if count > 0:
                self.log.info('Waiting on %s to become empty...' % queue.name)
                yield utils.tornado_sleep(sleep)
            else:
                self.log.debug('Queue is empty!')
                break

        raise gen.Return(True)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        raises: gen.Return()
        """
        pattern = self.option('name')
        matched_queues = yield self._fetch_queues(pattern)

        # Note: this does not check for dry mode.
        if self.option('required') and not matched_queues:
            raise QueueNotFound(
                'No queues like "%s" were found!' % pattern)

        self.log.info('Waiting for "%s" queues to become empty.' %
                      self.option('name'))

        sleepers = []
        for q in matched_queues:
            sleepers.append(self._wait(queue=q))

        self.log.info('%s queues need to be empty.' % len(matched_queues))
        self.log.info([q.name for q in matched_queues])
        yield sleepers
        self.log.info('All queues report empty.')

        raise gen.Return()



#!/usr/bin/env python
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc
"""CLI Script Runner for Kingpin."""

import argparse
import json
import logging
import os
import sys

from tornado import gen
from tornado import ioloop

from kingpin import utils
from kingpin.actors import utils as actor_utils
from kingpin.actors import exceptions as actor_exceptions
from kingpin.actors.misc import Macro
from kingpin.version import __version__


log = logging.getLogger(__name__)

__author__ = 'Matt Wise (matt@nextdoor.com)'


# We handle all the exceptions ourselves, so additional log statements from
# BOTO are not needed.
logging.getLogger('boto').setLevel(logging.CRITICAL)
logging.getLogger('botocore').setLevel(logging.CRITICAL)

# Initial option handler to set up the basic application environment.
parser = argparse.ArgumentParser(description='Kingpin v%s' % __version__)
parser.set_defaults(verbose=True)

# Job Configuration
parser.add_argument('-j', '--json', '-s', '--script', dest='script',
                    help='Path to JSON/YAML Deployment Script')
parser.add_argument('-a', '--actor', dest='actor',
                    help='Name of an Actor to execute (overrides --script)')
parser.add_argument('-E', '--explain', dest='explain', action='store_true',
                    help='Explain how an actor works. Requires --actor.',
                    default=False)
parser.add_argument('-p', '--param', dest='params', action='append',
                    help='Actor Parameter to set (ie, warn_on_failure=true)',
                    default=[])
parser.add_argument('-o', '--option', dest='options', action='append',
                    help='Actor Options to set (ie, elb_name=foobar)',
                    default=[])
parser.add_argument('-d', '--dry', dest='dry', action='store_true',
                    help='Executes a dry run only.')
parser.add_argument('--build-only', dest='build_only', action='store_true',
                    help='Compile the input JSON without executing any runs')
parser.add_argument('--orgchart', dest='orgchart',
                    help='Save the orgchart into file. Requires --build-only')

# Logging Configuration
parser.add_argument('-l', '--level', dest='level', default='info',
                    help='Set logging level (INFO|WARN|DEBUG|ERROR)')
parser.add_argument('-D', '--debug', dest='level_debug', default=False,
                    action='store_true', help='Equivalent to --level=DEBUG')
parser.add_argument('-c', '--color', dest='color', default=False,
                    action='store_true', help='Colorize the log output')

args = parser.parse_args()


def kingpin_fail(message):
    parser.print_help()
    sys.stderr.write('\nError: %s\n' % message)
    sys.exit(1)


def get_main_actor(dry):
    env_tokens = dict(os.environ)

    # Cannot specify a script file an an actor at the same time.
    if args.script and args.actor:
        kingpin_fail('You may only specify --actor or --script, not both!')

    if args.actor:
        ActorClass = actor_utils.get_actor_class(args.actor)
        parameters = dict([i.split('=') for i in args.params])
        options = dict([i.split('=') for i in args.options])

        return ActorClass(options=options,
                          dry=dry,
                          init_tokens=env_tokens,
                          **parameters)

    # Actor not specified. Process JSON file.
    try:
        script = args.script or sys.argv[1] if sys.argv else None
    except Exception as e:
        kingpin_fail(
            '%s You must specify --script or provide it as first argument.'
            % e)

    return Macro(desc='Kingpin',
                 options={'macro': script, 'tokens': env_tokens},
                 dry=dry)


@gen.coroutine
def main():

    if args.actor and args.explain:
        ActorClass = actor_utils.get_actor_class(args.actor)
        print((ActorClass.__doc__))
        sys.exit(0)

    if args.build_only:
        try:
            actor = get_main_actor(dry=False)
        except Exception as e:
            log.critical(e)
            sys.exit(1)

        if args.orgchart:
            log.info('Creating organizational chart into %s' % args.orgchart)
            try:
                orgdata = actor.get_orgchart()
            except Exception as e:
                log.critical(e)
                sys.exit(2)

            with open(args.orgchart, 'w') as output:
                output.write(json.dumps(orgdata))

        sys.exit(0)

    # Begin doing real stuff!
    if os.environ.get('SKIP_DRY', False):
        log.warning('')
        log.warning('*** You have disabled the dry run.')
        log.warning('*** Execution will begin with no expectation of success.')
        log.warning('')
    elif not args.dry:
        log.info('Rehearsing... Break a leg!')

        try:
            dry_actor = get_main_actor(dry=True)
            yield dry_actor.execute()
        except actor_exceptions.ActorException as e:
            log.critical('Dry run failed. Reason:')
            log.critical(e)
            sys.exit(2)

        log.info('Rehearsal OK! Performing!')

    try:
        runner = get_main_actor(dry=args.dry)

        log.info('')
        log.warning('Lights, camera ... action!')
        log.info('')
        yield runner.execute()
    except actor_exceptions.ActorException as e:
        log.error('Kingpin encountered mistakes during the play.')
        log.error(e)
        sys.exit(2)


def begin():
    # Set up logging before we do anything else
    if args.level_debug:
        args.level = 'DEBUG'
    utils.setup_root_logger(level=args.level, color=args.color)

    try:
        ioloop.IOLoop.instance().run_sync(main)
    except KeyboardInterrupt:
        log.info('CTRL-C Caught, shutting down')
        sys.exit(130)  # Standard KeyboardInterrupt exit code.
    except Exception as e:
        # Skip traceback that involves tornado's libraries.
        import traceback
        trace_lines = traceback.format_exc(e).splitlines()
        skip_next = False
        for l in trace_lines:
            if 'tornado' in l:
                skip_next = True
                continue
            if not skip_next:
                print(l)
            skip_next = False
        sys.exit(3)

if __name__ == '__main__':
    begin()



import logging
import time
import mock

from tornado import gen
from tornado import testing

from kingpin import utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors import group


log = logging.getLogger(__name__)


class TestActor(base.BaseActor):

    """Fake Actor for Tests"""

    all_options = {
        'value': (object, True, 'Intermediate value to be used'),
    }

    last_value = None

    @gen.coroutine
    def _execute(self):
        TestActor.last_value = self.option('value')
        raise gen.Return(None)


class TestActorRaises(base.BaseActor):

    """Fake Actor for Tests"""

    all_options = {
        'exception': (object, True, 'What this actor will return')
    }

    @gen.coroutine
    def _execute(self):
        exc = utils.str_to_class(self.option('exception'))
        raise exc


class TestActorPopulate(base.BaseActor):

    """Fake Actor for Tests"""

    all_options = {
        'object': (list, [], 'A list to append values.'),
        'value': (object, True, 'Add this value to the object two times.'),
    }

    last_value = None

    @gen.coroutine
    def _execute(self):
        self.option('object').append(self.option('value'))
        yield gen.moment
        self.option('object').append(self.option('value'))
        raise gen.Return(None)


class TestGroupActorBaseClass(testing.AsyncTestCase):

    def setUp(self, *args, **kwargs):
        super(TestGroupActorBaseClass, self).setUp(*args, **kwargs)
        TestActor.last_value = None
        self.actor_returns = {
            'desc': 'returns',
            'actor': 'kingpin.actors.test.test_group.TestActor',
            'options': {'value': None}}
        self.actor_with_a_problem = {
            'desc': 'Problematic',
            'actor': 'kingpin.actors.test.test_group.TestActor',
            'options': {'problem': 'unit-test-problem'}}
        self.actor_raises_unrecoverable_exception = {
            'desc': 'raises Unrecoverable exception',
            'actor': 'kingpin.actors.test.test_group.TestActorRaises',
            'options': {
                'exception':
                'kingpin.actors.exceptions.UnrecoverableActorFailure'
            }
        }
        self.actor_raises_recoverable_exception = {
            'desc': 'raises Recoverable exception',
            'actor': 'kingpin.actors.test.test_group.TestActorRaises',
            'options': {
                'exception':
                'kingpin.actors.exceptions.RecoverableActorFailure'
            }
        }


class TestBaseGroupActor(TestGroupActorBaseClass):

    def test_build_actions(self):
        actor = group.BaseGroupActor(
            'Unit Test Action',
            {'acts': [dict(self.actor_returns),
                      dict(self.actor_returns),
                      dict(self.actor_returns),
                      dict(self.actor_returns)]})
        ret = actor._build_actions()
        self.assertEqual(4, len(ret))

    def test_build_actions_with_bad_context_file(self):
        with self.assertRaises(exceptions.InvalidOptions):
            group.BaseGroupActor(
                'bad context',
                {'acts': [],
                 'contexts': {'file': 'no_such_file',
                              'tokens': {}}
                 }
            )

    def test_build_actions_with_context_file(self):
        acts = [dict(self.actor_returns)]

        with mock.patch.object(group.BaseGroupActor,
                               '_build_action_group') as action_builder:
            action_builder.return_value = acts
            group.BaseGroupActor(
                'ContextFile Actor',
                {
                    'acts': acts,
                    'contexts': {
                        'file': 'examples/test/context.json',
                        'tokens': {'TOKEN_VALUE': 'tadaa'}
                    }
                },
                init_context={'init': 'stuff'})

        self.assertEqual(2, len(action_builder.mock_calls))
        action_builder.assert_has_calls([
            mock.call(context={'init': 'stuff', 'key': 'value1'}),
            mock.call(context={'init': 'stuff', 'key': 'tadaa'})
        ])

    def test_build_actions_with_context_file_str(self):
        acts = [dict(self.actor_returns)]

        with mock.patch.object(group.BaseGroupActor,
                               '_build_action_group') as action_builder:
            action_builder.return_value = acts
            group.BaseGroupActor(
                'ContextFile Actor',
                {'acts': acts,
                 'contexts': 'examples/test/context.json'},
                init_tokens={'TOKEN_VALUE': 'tadaa'},
                init_context={'init': 'stuff'})

        self.assertEqual(2, len(action_builder.mock_calls))
        action_builder.assert_has_calls([
            mock.call(context={'init': 'stuff', 'key': 'value1'}),
            mock.call(context={'init': 'stuff', 'key': 'tadaa'})
        ])

    def test_build_actions_with_contexts(self):
        acts = [dict(self.actor_returns),
                dict(self.actor_returns),
                dict(self.actor_returns),
                dict(self.actor_returns)]

        with mock.patch.object(group.BaseGroupActor,
                               '_build_action_group') as action_builder:
            action_builder.return_value = acts
            group.BaseGroupActor(
                'Unit Test Action',
                {'acts': acts,
                 'contexts': [{'TEST': 'TestA'},
                              {'TEST': 'TestB'}]
                 },
                init_context={'PRE': 'CONTEXT'})

        self.assertEqual(2, len(action_builder.mock_calls))
        action_builder.assert_has_calls([
            mock.call(context={'PRE': 'CONTEXT', 'TEST': 'TestA'}),
            mock.call(context={'PRE': 'CONTEXT', 'TEST': 'TestB'})
        ])

    def test_build_action_group(self):
        acts = [dict(self.actor_returns),
                dict(self.actor_returns),
                dict(self.actor_returns),
                dict(self.actor_returns)]

        actor = group.BaseGroupActor('Unit Test Action', {'acts': acts})
        ret = actor._build_action_group({'TEST': 'CONTEXT'})
        self.assertEqual(ret[0]._init_context, {'TEST': 'CONTEXT'})

    @testing.gen_test
    def test_execute_success(self):
        actor = group.BaseGroupActor('Unit Test Action', {'acts': []})

        # Mock out the _run_actions method and make sure it just returns two
        # True results.
        @gen.coroutine
        def run_actions_true(*args, **kwargs):
            raise gen.Return([True, True])
        actor._run_actions = run_actions_true

        ret = yield actor._execute()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_execute_failure(self):
        actor = group.BaseGroupActor('Unit Test Action', {'acts': []})

        # Mock out the _run_actions method and make sure it just returns one
        # True and one False results.
        @gen.coroutine
        def run_actions_true(*args, **kwargs):
            raise gen.Return([True, False])
        actor._run_actions = run_actions_true

        ret = yield actor._execute()
        self.assertEqual(ret, None)


class TestSyncGroupActor(TestGroupActorBaseClass):

    @testing.gen_test
    def test_run_actions_with_no_acts(self):
        # Call the executor and test it out
        actor = group.Sync(
            'Unit Test Action', {'acts': []})

        res = yield actor._run_actions()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_run_actions_with_one_act(self):
        # Call the executor and test it out
        actor = group.Sync(
            'Unit Test Action',
            {'acts': [dict(self.actor_returns)]})

        res = yield actor._run_actions()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_run_actions_with_two_acts(self):
        # Call the executor and test it out
        actor = group.Sync(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_returns),
                dict(self.actor_returns)]})

        res = yield actor._run_actions()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_run_actions_continue_on_dry(self):
        # Call the executor and test it out
        self.actor_returns['options']['value'] = '123'
        actor = group.Sync(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_raises_unrecoverable_exception),
                dict(self.actor_returns),
            ]},
            dry=True)
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._run_actions()

        # Even after the first actor fails, the second one should get executed.
        self.assertEqual(TestActor.last_value, '123')

    @testing.gen_test
    def test_run_actions_with_two_acts_one_fails_unrecoverable(self):
        # Call the executor and test it out
        self.actor_returns['options']['value'] = '123'
        actor = group.Sync(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_raises_unrecoverable_exception),
                dict(self.actor_returns),
            ]})
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._run_actions()

        # If the second actor gets executed this value would be 123.
        self.assertEqual(TestActor.last_value, None)

    @testing.gen_test
    def test_run_actions_with_two_acts_one_fails_recoverable(self):
        # Call the executor and test it out
        actor = group.Sync(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_returns),
                dict(self.actor_raises_recoverable_exception)]})
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor._run_actions()


class TestAsyncGroupActor(TestGroupActorBaseClass):

    @testing.gen_test
    def test_get_exc_type_with_only_unrecoverable(self):
        exc_list = [
            exceptions.UnrecoverableActorFailure(),
            exceptions.UnrecoverableActorFailure(),
            exceptions.UnrecoverableActorFailure()
        ]
        actor = group.Async('Unit Test Action', {'acts': []})
        ret = actor._get_exc_type(exc_list)
        self.assertEqual(ret, exceptions.UnrecoverableActorFailure)

    @testing.gen_test
    def test_get_exc_type_with_only_recoverable(self):
        exc_list = [
            exceptions.RecoverableActorFailure(),
            exceptions.RecoverableActorFailure(),
            exceptions.RecoverableActorFailure()
        ]
        actor = group.Async('Unit Test Action', {'acts': []})
        ret = actor._get_exc_type(exc_list)
        self.assertEqual(ret, exceptions.RecoverableActorFailure)

    @testing.gen_test
    def test_get_exc_type_with_both(self):
        exc_list = [
            exceptions.RecoverableActorFailure(),
            exceptions.UnrecoverableActorFailure(),
            exceptions.RecoverableActorFailure()
        ]
        actor = group.Async('Unit Test Action', {'acts': []})
        ret = actor._get_exc_type(exc_list)
        self.assertEqual(ret, exceptions.UnrecoverableActorFailure)

    @testing.gen_test
    def test_run_actions_with_no_acts(self):
        # Call the executor and test it out
        actor = group.Async(
            'Unit Test Action', {'acts': []})

        res = yield actor._run_actions()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_run_actions_with_one_act(self):
        # Call the executor and test it out
        actor = group.Async(
            'Unit Test Action',
            {'acts': [dict(self.actor_returns)]})

        res = yield actor._run_actions()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_async(self):
        """Make sure this actor starts all processes in parallel!"""
        check_order = []
        actor_1 = {'actor': 'kingpin.actors.test.test_group.TestActorPopulate',
                   'desc': 'test',
                   'options': {
                       'object': ['fake'],
                       'value': 1}}
        actor_2 = {'actor': 'kingpin.actors.test.test_group.TestActorPopulate',
                   'desc': 'test',
                   'options': {
                       'object': ['fake'],
                       'value': 2}}
        actor = group.Async(
            'Unit Test Action',
            {
                'acts': [actor_1, actor_2]
            })

        # The options above were copied by value
        # This test requires same object to be modified so we set it directly
        actor._actions[0]._options['object'] = check_order
        actor._actions[1]._options['object'] = check_order

        yield actor.execute()

        # if the actions above were executed sequentially then the resulting
        # list would be [1,1,2,2] and here we know it's hopping between actors
        self.assertEqual(check_order, [1, 2, 1, 2])

    @testing.gen_test
    def test_execute_concurrent(self):
        sleeper = {'actor': 'misc.Sleep',
                   'desc': 'Sleep',
                   'options': {'sleep': 0.1}}
        actor = group.Async('Unit Test Action', {
            'concurrency': 2,
            'acts': [sleeper, sleeper, sleeper, sleeper]})

        start = time.time()
        yield actor.execute()
        stop = time.time()
        exe_time = stop - start
        self.assertTrue(0.2 < exe_time < 0.4)

    @testing.gen_test
    def test_run_actions_with_two_acts(self):
        # Call the executor and test it out
        actor = group.Async(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_returns),
                dict(self.actor_returns)]})

        res = yield actor._run_actions()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_run_actions_with_two_acts_one_fails_unrecoverable(self):
        # Call the executor and test it out
        self.actor_returns['options']['value'] = '123'
        actor = group.Async(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_raises_unrecoverable_exception),
                dict(self.actor_returns),
            ]})

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._run_actions()

        # If the second actor does not get executed this value would be None
        self.assertEqual(TestActor.last_value, '123')

    @testing.gen_test
    def test_run_actions_with_two_acts_one_fails_recoverable(self):
        # Call the executor and test it out
        actor = group.Async(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_returns),
                dict(self.actor_raises_recoverable_exception),
                dict(self.actor_raises_recoverable_exception)]})

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor._run_actions()

    @testing.gen_test
    def test_run_actions_with_two_acts_one_fails_with_both(self):
        # Call the executor and test it out
        actor = group.Async(
            'Unit Test Action',
            {'acts': [
                dict(self.actor_returns),
                dict(self.actor_raises_recoverable_exception),
                dict(self.actor_raises_unrecoverable_exception)]})

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._run_actions()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.ecs`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import jsonschema
import logging
import operator

from tornado import gen

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.aws import settings
from kingpin.actors.utils import dry
from kingpin.constants import REQUIRED, STATE

log = logging.getLogger(__name__)

__author__ = 'Willem Van Eck <willem@nextdoor.com>'


class ECSAPIException(exceptions.RecoverableActorFailure):
    """A failure from the ECS API."""


class ECSTaskFailedException(exceptions.RecoverableActorFailure):
    """A failure from an ECS Task."""


class ServiceNotFound(exceptions.RecoverableActorFailure):
    """Failure to find an ECS Service."""


# http://boto3.readthedocs.io/en/latest/reference/services/ecs.html
TASK_DEFINITION_SCHEMA = {
    'type': 'object',
    'required': ['family', 'containerDefinitions'],
    'properties': {
        'family': {'type': 'string'},
        'networkMode': {'type': 'string'},
        'containerDefinitions': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {'type': 'string'},
                    'image': {'type': 'string'},
                    'cpu': {'type': 'number'},
                    'memory': {'type': 'number'},
                    'memoryReservation': {'type': 'number'},
                    'links': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'portMappings': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'containerPort': {'type': 'number'},
                                'hostPort': {'type': 'number'},
                                'protocol': {'type': 'string'}
                            }
                        }
                    },
                    'essential': {'type': 'boolean'},
                    'entryPoint': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'command': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'environment': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'name': {'type': 'string'},
                                'value': {'type': 'string'}
                            }
                        }
                    },
                    'mountPoints': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'sourceVolume': {'type': 'string'},
                                'containerPath': {'type': 'string'},
                                'readOnly': {'type': 'boolean'}
                            }
                        }
                    },
                    'systemControls': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'namespace': {'type': 'string'},
                                'value': {'type': 'string'}
                            }
                        }
                    },
                    'volumesFrom': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'sourceContainer': {'type': 'string'},
                                'readOnly': {'type': 'boolean'}
                            }
                        }
                    },
                    'hostname': {'type': 'string'},
                    'user': {'type': 'string'},
                    'workingDirectory': {'type': 'string'},
                    'disableNetworking': {'type': 'boolean'},
                    'privileged': {'type': 'boolean'},
                    'readonlyRootFilesystem': {'type': 'boolean'},
                    'dnsServers': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'dnsSearchDomains': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'extraHosts': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'properties': {
                                'hostname': {'type': 'string'},
                                'ipAddress': {'type': 'string'}
                            }
                        }
                    },
                    'dockerSecurityOptions': {
                        'type': 'array',
                        'items': {'type': 'string'}
                    },
                    'dockerLabels': {'type': 'object'},
                    'ulimits': {
                        'type': 'array',
                        'items': {
                            'type': 'object',
                            'required': ['name', 'softLimit', 'hardLimit'],
                            'properties': {
                                'name': {'type': 'string'},
                                'softLimit': {'type': 'number'},
                                'hardLimit': {'type': 'number'}
                            }
                        }
                    },
                    'logConfiguration': {'type': 'object'},
                    'logDriver': {'type': 'string'},
                    'options': {'type': 'object'}
                }
            }
        },
        'volumes': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {'type': 'string'},
                    'host': {
                        'type': 'object',
                        'properties': {
                            'sourcePath': {'type': 'string'}
                        }
                    }
                }
            }
        },
        'tags': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'key': {'type': 'string'},
                    'value': {'type': 'string'}
                }
            }
        }
    }
}


class ECSBaseActor(base.AWSBaseActor):
    """Base class for ECS actors."""

    all_options = {
        'count': ((int, str), 1, 'How many tasks to run.')
    }

    def __init__(self, *args, **kwargs):
        super(ECSBaseActor, self).__init__(*args, **kwargs)

        count = self.option('count')
        if type(count) is str or type(count) is str:
            try:
                self._options['count'] = int(count)
            except ValueError:
                raise exceptions.RecoverableActorFailure(
                    'Could not parse option \'count\' as int: %s' % count)

    FAILURE_MISSING = 'MISSING'

    def _handle_failures(self, failures, *ignorable):
        """Logs ECS failures. Raises a recoverable exception if there are any.

        Args:
            failures: list of failures from ECS api.
            ignorable: list of failures to ignore.
        """
        if not failures:
            return

        failure_reasons = []

        for failure in failures:
            reason = failure['reason']
            if reason in ignorable:
                continue
            failure_reasons.append(reason)
            self.log.error(failure)
        if failure_reasons:
            raise ECSAPIException(failure_reasons)

    @gen.coroutine
    @dry('Would register task definition with family {0[family]}')
    def _register_task(self, task_definition):
        """Registers a task.

        Args:
            task_definition: dict of ECS task definition parameters.

        Returns:
            Task Definition name string.
        """
        family = task_definition['family']

        self.log.info('Registering task definition with family {}'.format(
            family))

        response = yield self.api_call(
            self.ecs_conn.register_task_definition, **task_definition)

        # Parse data from the server's response.
        task_definition = response['taskDefinition']
        revision = task_definition['revision']

        task_definition_name = '{}:{}'.format(family, revision)
        self.log.info('Task definition {} registered.'.format(
            task_definition_name))
        raise gen.Return(task_definition_name)

    @gen.coroutine
    @dry('Would deregister task definition {0}')
    def _deregister_task_definition(self, task_definition_name):
        """Deregisters a task definition.

        Args:
            task_definition_name: Task Definition name or arn to deregister.
        """
        self.log.info(
            'Deregistering task definition {}.'.format(task_definition_name))
        yield self.api_call(
            self.ecs_conn.deregister_task_definition,
            taskDefinition=task_definition_name)

    def _read_list_task_definitions_paginator(self, **kwargs):
        """Reads and aggregates results from a list_task_definitions paginator.

        Args:
            **kwargs: Args to pass to the paginator.

        Returns:
            List of task definition arns.
        """
        paginator = self.ecs_conn.get_paginator('list_task_definitions')
        task_definitions = []
        i = 0
        page_iterator = paginator.paginate(**kwargs)
        for page in page_iterator:
            self.log.info('Parsing page {} of results.'.format(i))
            i += 1
            task_definitions += page['taskDefinitionArns']

        return task_definitions

    @gen.coroutine
    @dry('Would describe task definition {0}')
    def _describe_task_definition(self, task_definition_name):
        """Describes the given task definition.

        Args:
            task_definition_name: Task Definition string.

        Returns:
            Returns a dict containing the task definition dict and task
            definition tags list.
        """
        self.log.info('Describing task definition {}.'.format(
            task_definition_name))
        resp = yield self.api_call(
            self.ecs_conn.describe_task_definition,
            taskDefinition=task_definition_name,
            include=['TAGS'])

        del resp['ResponseMetadata']

        raise gen.Return(resp)

    @gen.coroutine
    @dry('Would list task definitions')
    def _list_task_definitions(self, status='ALL', family_prefix=None):
        """List task definitions.

        Optionally uses a prefix to filter family names.
        Optionally returns active task definitions.

        Args:
            status: Type of status to filter with.
                Valid types are 'ACTIVE', 'INACTIVE', and 'ALL'.
                Default 'ALL'.
            family_prefix: Prefix to filter task definition families with.
                Default None.

        Returns:
            List of task definitions arns matching specified restrictions.
        """
        self.log.info(
            'Listing task definitions '
            'with status {} and family prefix {}.'.format(
                status, family_prefix))
        task_definitions = yield self.api_call(
            self._read_list_task_definitions_paginator,
            status=status,
            familyPrefix=family_prefix)

        raise gen.Return(task_definitions)

    @staticmethod
    def _load_task_definition(task_definition_file, tokens, default_tokens={}):
        """Loads and verifies a task definition template file, interpolates
        tokens, and optionally default tokens which may contain environment
        variables.

        Args:
            task_definition_file: task definition file to load, or None.
            tokens: dict of key/value pairs to interpolate into the file.
            default_tokens: dict of default key/value pairs to merge
                with tokens

        Returns:
            Resulting task definition dict or
                None if task_definition_file is None.
        """
        if not task_definition_file:
            return None

        # Defined Kingpin tokens will override environment variables.
        final_tokens = default_tokens.copy()
        final_tokens.update(tokens)

        task_definition = utils.convert_script_to_dict(
            task_definition_file, final_tokens)

        try:
            jsonschema.validate(task_definition,
                                TASK_DEFINITION_SCHEMA)
        except jsonschema.exceptions.ValidationError as e:
            raise exceptions.InvalidOptions(e)
        return task_definition

    @staticmethod
    def _load_service_definition(
            service_definition_file, tokens, default_tokens=None):
        """Loads and verifies a service definition template file, and
        interpolates tokens. and optionally default tokens which may contain
        environment variables. The service definition template file
        can be None.

        Args:
            service_definition_file: service definition file to load.
                If None or an empty string, this returns only defaults.
            tokens: dict of key/value pairs to interpolate into the file.
            default_tokens: dict of default key/value pairs to merge
                with tokens

        Returns:
            Resulting service definition dict.
        """
        if default_tokens is None:
            default_tokens = {}
        if not service_definition_file:
            service_definition = {}
        else:
            final_tokens = default_tokens.copy()
            final_tokens.update(tokens)

            service_definition = utils.convert_script_to_dict(
                service_definition_file, final_tokens)
            try:
                jsonschema.validate(service_definition,
                                    SERVICE_DEFINITION_SCHEMA)

            except jsonschema.exceptions.ValidationError as e:
                raise exceptions.InvalidOptions(e)

        return service_definition

    @staticmethod
    def _arn_to_name(arn):
        """Gets the name in an arn.

        Example input:
            arn:aws:ecs:region:account-id:task-definition/name:135
        Example output:
            name:135

        Args:
            arn: The arn to process.

        Returns:
            The name.
        """
        return arn[arn.index('/') + 1:]

    def _format_location(self):
        """Get a formatted location describing where this actor is running
        for logging.

        Returns:
            Formatted location string with region and cluster.
        """
        return "'{}', cluster '{}'".format(
            self.option('region'),
            self.option('cluster'))


class RunTask(ECSBaseActor):
    """Register and run a task on ECS.

    This actor will loop indefinitely until the task is complete.

    **Options**

    :region:
      AWS region (or zone) name, such as us-west-2 or eu-west-1.

    :cluster:
      ECS cluster in the region to launch the task into.

    :task_definition:
      String of path to the Task Definition file template.
      Must be a local file path.
      Tokens to be interpolated must be of the form %VAR%.
      Tokens can come in the form of options or environment variables.

    :tokens:
      A dictionary of key/value pairs used to fill in the tokens for the
      Task Definition template.
      These will override environment variables which can be used as tokens.
      Default: {}.

    :count:
      How many tasks to run. Default: 1.

    :wait:
      Whether to wait for the tasks to complete. Default: True.

    **Examples**

    .. code-block:: yaml

       actor: aws.ecs.Task
       desc: Run migrations
       options:
          task_definition: migrate.yaml
          region: us-west-2
          cluster: us1-internal

    **Dry Mode**

    Will only attempt to interpolate env vars into the task definition.
    """

    all_options = {
        'region': (str, REQUIRED,
                   'AWS region (or zone) name, like us-west-2 or eu-west-1.'),
        'cluster': (str, REQUIRED,
                    'ECS cluster in the region to launch the task into.'),
        'task_definition':
            (str, REQUIRED,
             'String of path to the Task Definition file template. '
             'Must be a local file path. '
             'Tokens to be interpolated must be of the form %VAR%.'),
        'tokens': (dict, {},
                   'A dictionary of key/value pairs '
                   'used to fill in the tokens for the '
                   'Task Definition template.'),
        'count': ((int, str), 1, 'How many tasks to run.'),
        'wait': (bool, True,
                 'Whether to wait for the tasks to complete.')
    }

    def __init__(self, *args, **kwargs):
        super(RunTask, self).__init__(*args, **kwargs)
        self.task_definition = self._load_task_definition(
            self.option('task_definition'),
            self.option('tokens'),
            self._init_tokens)

    @gen.coroutine
    @dry('Would run task {0}')
    def _run_task(self, task_definition_name):
        """Runs a task on ECS given a task definition's family and revision.

        A task can result in multiple running tasks,
        depending on count and sidekick tasks.

        Args:
            task_definition_name: Task Definition string

        Returns:
            list: task ARNs.
        """
        repeating_log = utils.create_repeating_log(
            self.log.info,
            'Waiting for task to be found...',
            seconds=30)

        while True:
            response = yield self.api_call(
                self.ecs_conn.run_task,
                cluster=self.option('cluster'),
                taskDefinition=task_definition_name,
                count=self.option('count'))

            if not response['failures']:
                break
            # Error on non-missing failures.
            self._handle_failures(response['failures'], self.FAILURE_MISSING)
            yield gen.sleep(2)

        utils.clear_repeating_log(repeating_log)

        self.log.info('Scheduled task {}.'.format(task_definition_name))
        tasks = [t['taskArn'] for t in response['tasks']]
        raise gen.Return(tasks)

    @gen.coroutine
    def _wait_for_tasks(self, tasks):
        """Wait for tasks to complete on ECS.

        If a task stops with a non-zero exit code,
        this will raise an exception.

        Args:
            tasks: list of task ARNs to wait for.
        """
        if not tasks:
            return
        while True:
            done = yield self._tasks_done(tasks)
            if done:
                break
            yield gen.sleep(10)

    @gen.coroutine
    @utils.retry(excs=ECSAPIException,
                 retries=settings.ECS_RETRY_ATTEMPTS,
                 delay=settings.ECS_RETRY_DELAY)
    def _tasks_done(self, tasks):
        """Checks if tasks are done.

        Args:
            tasks: list of task ARNs to check.

        Returns:
            A boolean indicating whether all tasks are done.
        """
        response = yield self.api_call(
            self.ecs_conn.describe_tasks,
            cluster=self.option('cluster'),
            tasks=tasks)

        self._handle_failures(response['failures'])

        containers = self._get_containers_from_tasks(
            task_list=response['tasks'])

        total_count = len(containers)
        stopped_count = 0

        for container in containers:
            if container['lastStatus'] == 'STOPPED':
                stopped_count += 1
                task_id = container['taskArn']
                if 'reason' in container:
                    self.log.warning('Error reason for {}: {}.'.format(
                        task_id, container['reason']))
                exit_code = container.get('exitCode', None)
                if exit_code is None:
                    raise ECSTaskFailedException(
                        'Task {} stopped without executing.'.format(
                            task_id))
                if exit_code != 0:
                    raise ECSTaskFailedException(
                        'Task {} errored out with exit code {}.'.format(
                            task_id, exit_code))
                self.log.info('Task {} finished successfully!'.format(
                    task_id))

        if stopped_count == total_count:
            self.log.info('All {} tasks finished.'.format(total_count))
            raise gen.Return(True)

        self.log.info(
            '{stopped} tasks finished out of {total}.'.format(
                stopped=stopped_count,
                total=total_count))
        raise gen.Return(False)

    @staticmethod
    def _get_containers_from_tasks(task_list):
        """Helper function to get a flat list of containers out of a list of tasks.

        Args:
            task_list: List of dictionaries with the format:
                [
                    {
                        ...
                        'containers': [ ... ]
                        ...
                    }
                ]

        Returns:
            All containers as a flat list.
        """
        containers = []
        for task_overview in task_list:
            containers.extend(task_overview['containers'])
        return containers

    @gen.coroutine
    def _execute(self):
        self.log.info(
            'Running task from {task_definition} in ECS. '
            'Region: {region}, cluster: {cluster}'.format(
                task_definition=self.option('task_definition'),
                region=self.option('region'),
                cluster=self.option('cluster')))
        task_definition_name = yield self._register_task(
            self.task_definition)

        tasks = yield self._run_task(task_definition_name)
        if self.option('wait'):
            yield self._wait_for_tasks(tasks)
        else:
            self.log.info('Not waiting for tasks to complete.')


# http://boto3.readthedocs.io/en/latest/reference/services/ecs.html
SERVICE_DEFINITION_SCHEMA = {
    'type': 'object',
    'properties': {
        'loadBalancers': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'loadBalancerName': {'type': 'string'},
                    'containerName': {'type': 'string'},
                    'containerPort': {'type': 'number'},
                }
            }
        },
        'role': {'type': 'string'},
        'deploymentConfiguration': {
            'type': 'object',
            'properties': {
                'maximumPercent': {'type': 'number'},
                'minimumHealthyPercent': {'type': 'number'},
            }
        }
    }
}


class Service(ECSBaseActor):
    """Register and run a service on ECS.

    This actor will loop indefinitely until the task is complete. If any
    Service parameters are not supplied, then Amazon supplies the defaults and
    manages them. If these are immutable in Amazon, then you cannot change them
    in the ECS Service Definition down in a future update, and Kingpin will
    error out.

    If the service already exists, it is upgraded.

    **Options**

    :state:
      Desired state: present/absent

    :region:
      AWS region (or zone) name, such as us-west-2 or eu-west-1.

    :cluster:
      ECS cluster in the region to launch the task into.

    :task_definition:
      String of path to the Task Definition file template.
      Must be a local file path.
      Tokens to be interpolated must be of the form %VAR%.
      Tokens can come in the form of options or environment variables.
      Default: None.

    :service_definition:
      String of path to the Service Definition file template.
      Must be a local file path.
      Tokens to be interpolated must be of the form %VAR%.
      Implicit fields - do not include these:
      'serviceName', 'taskDefinition', 'desiredCount', 'clientToken'
      Default: None.

    :service_name:
      Service name to use. If not specified, this will use
      the Task Definition's family.
      Necessary if there is no given Task Definition.
      Default: None.

    :tokens:
      A dictionary of key/value pairs used to fill in the tokens for the
      Task and Service Definition template.
      These will override environment variables which can be used as tokens.
      Default: {}.

    :count:
      How many instances of the service to deploy.
      Not used when state is 'absent'.
      Default: 1.

    :wait:
      Whether to wait for the services to deploy.
      Not used when state is 'absent'.
      Default: True.

    :use_existing_count:
      Whether to use the existing service's count when updating
      instead of the specified count.
      Default: True.

    :deregister_task_definitions:
      Whether to deregister related Task Definitions.
      Only used when state is 'absent'.

    **Examples**

    .. code-block:: yaml

       actor: aws.ecs.Service
       desc: Deploy taskworker
       options:
          task_definition: taskworker.yaml
          region: us-west-2
          cluster: us1-internal

    **Dry Mode**

    Will only validate and interpolate tokens into both the
    task and service definition files.
    """

    all_options = {
        'state': (STATE, 'present', 'Desired state: present/absent'),
        'region': (str, REQUIRED,
                   'AWS region (or zone) name, like us-west-2 or eu-west-1.'),
        'cluster': (str, REQUIRED,
                    'ECS cluster in the region to launch the task into.'),
        'task_definition':
            (str, None,
             'String of path to the Task Definition file template. '
             'Must be a local file path. '
             'Tokens to be interpolated must be of the form %VAR%.'),
        'service_definition':
            (str, None,
             'String of path to the Service Definition file template. '
             'Must be a local file path. '
             'Tokens to be interpolated must be of the form %VAR%. '
             'Implicit fields - do not include these: '
             "'serviceName', 'taskDefinition', 'desiredCount', 'clientToken'"),
        'service_name':
            (str, None,
             'Service name to use. If not specified, this will use '
             "the Task Definition's family."
             'Necessary if there is no given Task Definition.'),
        'tokens': (dict, {},
                   'A dictionary of key/value pairs '
                   'used to fill in the tokens for the Task and Service '
                   'definition templates.'),
        'count': ((int, str), 1,
                  'How many instances of the service to deploy. '
                  "Not used when state is 'absent'."),
        'wait': (bool, True,
                 'Whether to wait for the services to deploy. '
                 "Not used when state is 'absent'."),
        'use_existing_count': (
            bool, True,
            "Whether to use the existing service's count when updating "
            'instead of the specified count.'),
        'deregister_task_definitions': (
            bool, True,
            'Whether to deregister related Task Definitions. '
            "Only used when state is 'absent'.")
    }

    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)
        self.task_definition = self._load_task_definition(
            self.option('task_definition'),
            self.option('tokens'),
            self._init_tokens)
        self.service_definition = self._load_service_definition(
            self.option('service_definition'),
            self.option('tokens'),
            self._init_tokens)

    @gen.coroutine
    def _describe_service(self, service_name):
        """Describe a service by name.

        Args:
            service_name: service name to describe.

        Returns:
            Service description dict.

        Raises:
            RecoverableActorFailure if number of services found is not 1.
        """
        response = yield self.api_call(
            self.ecs_conn.describe_services,
            cluster=self.option('cluster'),
            services=[service_name])
        self._handle_failures(response['failures'], self.FAILURE_MISSING)

        services = response['services']

        # There should never be more than one service for a given name.
        if len(services) != 1:
            raise ServiceNotFound(
                'API returned {count} services for {name}'.format(
                    count=len(services), name=service_name))

        raise gen.Return(services[0])

    @staticmethod
    def _get_primary_deployment(service):
        """Gets the primary deployment from a service.

        Args:
            service: Result of _describe_service
                to get the primary deployment from.

        Returns:
            Primary deployment dict, or None if there is no primary deployment.
        """
        deployments = service['deployments']
        primary_deployment = None
        for deployment in deployments:
            if deployment['status'] == 'PRIMARY':
                primary_deployment = deployment
                break
        return primary_deployment

    @gen.coroutine
    def _wait_for_deployment_update(self, service_name, task_definition_name):
        """Wait's for a service's primary deployment to be updated.

        Args:
            service_name: Service name to wait for.
            task_definition_name: Expected Task Definition string.
        """
        repeating_log = utils.create_repeating_log(
            self.log.info,
            'Waiting for primary deployment to be updated to %s '
            'for service with name %s...' % (task_definition_name,
                                             service_name),
            seconds=30)

        while True:
            try:
                service = yield self._describe_service(service_name)
            except ServiceNotFound as e:
                self.log.info('Service Not Found: %s' % str(e))
                yield gen.sleep(2)
                continue

            primary_deployment = self._get_primary_deployment(service)
            if primary_deployment:
                self.log.info('Primary deployment is %s.' %
                              self._arn_to_name(
                                  primary_deployment['taskDefinition']))
                if self._is_task_in_deployment(
                        primary_deployment, task_definition_name):
                    self.log.info('Primary deployment updated.')
                    break
            yield gen.sleep(2)

        utils.clear_repeating_log(repeating_log)

    @gen.coroutine
    def _is_task_definition_different(self, old_task_definition_name,
                                      new_task_definition_name):
        """Checks if the actor's task definition is different
        than the given service's.

        Args:
            old_task_definition_name: Task Definition string of the
                old task definition.
            new_task_definition_name: Task Definition string of the
                new task definition.

        Returns:
            A boolean which is true if there are no differences.
        """
        old_task_definition = yield self._describe_task_definition(
            old_task_definition_name)
        new_task_definition = yield self._describe_task_definition(
            new_task_definition_name)

        del new_task_definition['taskDefinition']['revision']
        del new_task_definition['taskDefinition']['taskDefinitionArn']
        del old_task_definition['taskDefinition']['revision']
        del old_task_definition['taskDefinition']['taskDefinitionArn']

        raise gen.Return(old_task_definition != new_task_definition)

    @gen.coroutine
    @dry('Would create service')
    def _create_service(self, service_name):
        """Create a service.

        Args:
            service_name: Service name to use.

        Returns:
            task_definition_name: Task Definition string that was registered.
        """
        # Need to register a new task definition and deploy.
        task_definition_name = yield self._register_task(
            self.task_definition)

        create_parameters = dict(
            cluster=self.option('cluster'),
            serviceName=service_name,
            taskDefinition=task_definition_name,
            desiredCount=self.option('count'),
            **self.service_definition)

        self.log.info('Creating service.')

        yield self.api_call(
            self.ecs_conn.create_service,
            **create_parameters)

        yield self._wait_for_deployment_update(service_name,
                                               task_definition_name)
        self.log.info('Finished creating service.')
        raise gen.Return(task_definition_name)

    @gen.coroutine
    @dry('Would update service')
    def _update_service(self, service_name, existing_service, override=None):
        """Update a service.

        Args:
            service_name: Service name to use.
            existing_service: Result of _describe_service
                on the existing service.
            override: List of tuples of key:value to override.

        Returns:
            Task Definition string that was updated or registered.
        """
        update_parameters = dict(
            cluster=self.option('cluster'),
            service=service_name,
            **self.service_definition)

        # Remove any fields that aren't allowed on update.
        for immutable_field in self._immutable_fields():
            update_parameters.pop(immutable_field, None)

        old_task_definition_name = self._arn_to_name(
            existing_service['taskDefinition'])
        task_definition_name = old_task_definition_name

        is_new_task_definition = False
        if self.task_definition:
            new_task_definition_name = yield self._register_task(
                self.task_definition)
            is_new_task_definition = yield self._is_task_definition_different(
                old_task_definition_name,
                new_task_definition_name)
            if is_new_task_definition:
                self.log.info(
                    'The task definition is different, using the new one.')
                # Need to use the new task definition.
                update_parameters.update({
                    'taskDefinition': new_task_definition_name})
                task_definition_name = new_task_definition_name
            else:
                self.log.info(
                    'The task definition is the same, using the old one.')
                yield self._deregister_task_definition(
                    new_task_definition_name)
                task_definition_name = old_task_definition_name

        if override is not None:
            update_parameters.update(override)

        self.log.info('Updating service.')
        self.log.info('Checking if existing service is active...')
        service_is_active = (existing_service and
                             existing_service['status'] != 'INACTIVE')
        if not service_is_active:
            # We can only update an existing, active service.
            self.log.error(
                'Could not find service with name {} to update '
                'in {}. Update is likely to fail!'.format(
                    service_name, self._format_location()))
            current_primary_deployment = self._get_primary_deployment(
                existing_service)
            self.log.info(
                'Current primary deployment task definition is: {}'.format(
                    current_primary_deployment['taskDefinition']))
            current_service_description = str(self._describe_service(
                service_name))
            self.log.info(
                'Service description is: {}'.format(
                    current_service_description))
        yield self.api_call(
            self.ecs_conn.update_service,
            **update_parameters)

        if is_new_task_definition:
            yield self._wait_for_deployment_update(service_name,
                                                   task_definition_name)
        self.log.info('Finished updating service.')
        raise gen.Return(task_definition_name)

    @gen.coroutine
    @dry('Would stop service')
    def _stop_service(self, service_name, existing_service):
        """Stop all the tasks in a service.

        Args:
            service_name: name of the service to stop.
            existing_service: result of _describe_service on the given
                service name.
        """
        self.log.info('Shutting down all current tasks in %s.' % service_name)
        task_definition_name = yield self._update_service(
            service_name,
            existing_service,
            override={'desiredCount': 0})
        yield self._wait_for_service_update(service_name, task_definition_name)
        self.log.info('Service {} stopped successfully.'.format(service_name))

    @gen.coroutine
    @dry('Would delete service')
    def _delete_service(self, service_name, existing_service,
                        deregister=True):
        """Delete a service.

        This also deregisters task definitions with the same family
        unless deregister is False.

        Args:
            service_name: name of the service to delete.
            existing_service: result of _describe_service on the given
                service name.
            deregister: Whether to deregister related Task Definitions.
        """
        status = existing_service['status']
        if status != 'ACTIVE':
            # Either DRAINING or INACTIVE
            self.log.info(
                'Service {} is already not active, is {}.'.format(
                    service_name, status))
        else:
            yield self._stop_service(service_name, existing_service)
            yield self.api_call(self.ecs_conn.delete_service,
                                cluster=self.option('cluster'),
                                service=service_name)
        if deregister:
            task_definition_names = yield self._list_task_definitions(
                status='ACTIVE',
                family_prefix=service_name)
            for task_definition_name in task_definition_names:
                yield self._deregister_task_definition(task_definition_name)

    def _is_task_in_deployment(self, deployment, task_definition_name):
        """Checks whether the given task definition is in the deployment.
        This is useful for checking that we're looking at the right deployment.

        Args:
            deployment: Deployment to check.
            task_definition_name: Task Definition string to look for.

        Returns:
            A boolean, true if the deployment has the task definition.
        """
        return task_definition_name == self._arn_to_name(
            deployment['taskDefinition'])

    def _check_immutable_field_errors(self, old_params, new_params,
                                      immutable_fields):
        """Compares an old service definition to a new one
         to ensure that all of the specified immutable fields
         are the same between them.

        If there are any errors,
        this logs them and raises RecoverableActorFailure.

        Args:
            old_params: old parameters to use.
            new_params: new parameters to compare with.
            immutable_fields: list of immutable fields.
        """
        # API does not return role name, only the role arn.
        role_arn = old_params.get('roleArn')
        role = None
        if role_arn:
            role = self._arn_to_name(role_arn)
        old_params['role'] = role

        has_error = False
        for immutable_field_name in immutable_fields:
            new_field = new_params.get(immutable_field_name)
            old_field = old_params.get(immutable_field_name)

            # If the newly supplied fields are None, then they aren't defined
            # in the service definition at all. It's OK to ignore them.
            if new_field is None:
                continue

            # If the supplied fields are an empty list, treat them also as an
            # null field and move on.
            if isinstance(new_field, list) and len(new_field) == 0:
                continue

            if new_field != old_field:
                has_error = True
                self.log.error(
                    "Field \'{field}\' cannot be updated.\n"
                    'Old value: {old}\n'
                    'New value: {new}'.format(
                        field=immutable_field_name,
                        old=old_field,
                        new=new_field))

        if has_error:
            raise exceptions.RecoverableActorFailure(
                'Immutable fields cannot be updated. '
                'A new service must be created.')

    def _immutable_fields(self):
        """
        Return: List of all fields that ECS doesn't allow mutation on.
        """
        return [
            'loadBalancers', 'launchType', 'role',
            'placementConstraints', 'placementStrategy']

    @gen.coroutine
    @dry('Would wait for service to update its state successfully')
    def _wait_for_service_update(self, service_name, task_definition_name):
        """Wait for the service's state to successfully update.

        If the service fails to be updated for some reason,
        e.g. failing to come up during deployment,
        this will raise an exception.

        Args:
            service_name: name of the service to wait for.
            task_definition_name: Task Definition string.
        """
        # Create set used to ensure event logs are only printed once.
        self.seen_events = set()
        while True:
            done = yield self._is_service_updated(
                service_name, task_definition_name)
            if done:
                break
            yield gen.sleep(10)

    @gen.coroutine
    @utils.retry(excs=exceptions.RecoverableActorFailure,
                 retries=settings.ECS_RETRY_ATTEMPTS,
                 delay=settings.ECS_RETRY_DELAY)
    def _is_service_updated(self, service_name, task_definition_name):
        """Checks if service's state updates successfully.
        Meant to be called in a wait-loop.

        Args:
            service_name: name of the service to wait for.
            task_definition_name: Task Definition string.

        Returns:
            A boolean indicating whether the service is completely updated.
        """
        service = yield self._describe_service(service_name)

        deployments = service['deployments']
        primary_deployment = self._get_primary_deployment(service)
        if not primary_deployment:
            # There should always be one 'PRIMARY' deployment returned.
            raise exceptions.RecoverableActorFailure(
                'No primary deployment.')

        # Verify that the primary deployment has the correct task definition.
        if not self._is_task_in_deployment(
                primary_deployment, task_definition_name):
            raise exceptions.RecoverableActorFailure(
                'Primary deployment was for {}, not {}.'.format(
                    self._arn_to_name(primary_deployment['taskDefinition']),
                    task_definition_name))

        service_timestamp = primary_deployment['updatedAt']

        sorted_new_events = self._get_sorted_new_log_events(
            events=service['events'],
            start_timestamp=service_timestamp)

        for event in sorted_new_events:
            event_timestamp, event_message = event
            self.log.info('Event [{}]: {}'.format(event_timestamp,
                                                  event_message))

        running_count = primary_deployment['runningCount']
        desired_count = primary_deployment['desiredCount']
        missing_count = desired_count - running_count

        extra_deployment_count = len(deployments) - 1

        if missing_count == 0 and extra_deployment_count == 0:
            raise gen.Return(True)

        self.log.info(
            '{} tasks running out of {}, '
            'and {} deployments waiting on termination.'.format(
                running_count, desired_count, extra_deployment_count))
        raise gen.Return(False)

    def _get_sorted_new_log_events(self, events, start_timestamp):
        """Retrieves a list of sorted, new log events.

        Only logs events which:
            Have timestamps greater than start_timestamp.
            Have not been logged before (based on self.seen_events).

        Args:
            events: list of event dicts.
                Must have fields 'id', 'message', and 'createdAt'.
            start_timestamp: timestamp before which to ignore events.

        Returns:
            A list of event tuples like [(event_timestamp, event_message)...]
        """
        # Get list of all new events.
        new_events = []
        for event in events:
            event_id = event['id']
            event_message = event['message']
            event_timestamp = event['createdAt']
            if event_timestamp < start_timestamp:
                continue
            if event_id in self.seen_events:
                continue
            self.seen_events.add(event_id)
            new_events.append((event_timestamp, event_message))

        # Sort events on timestamp.
        return sorted(new_events, key=operator.itemgetter(0))

    @gen.coroutine
    @dry('Would ensure the service is absent')
    def _ensure_service_absent(self, service_name, existing_service):
        """Ensures a service is absent.

        Args:
            service_name: service name to use.
            existing_service:
                Result of _describe_service on the given service name.
        """
        if existing_service:
            yield self._delete_service(
                service_name, existing_service,
                deregister=self.option('deregister_task_definitions'))
            self.log.info(
                'Service {} deleted successfully in {}.'.format(
                    service_name, self._format_location()))
        else:
            self.log.info(
                'Service {} already absent in {}.'.format(
                    service_name, self._format_location()))

    @gen.coroutine
    @dry('Would ensure the service is present')
    def _ensure_service_present(self, service_name, existing_service):
        """Ensures a service is present.

        Args:
            service_name: service name to use.
            existing_service:
                Result of _describe_service on the given service name.
        """
        service_is_active = (existing_service and
                             existing_service['status'] != 'INACTIVE')
        if not service_is_active and not self.task_definition:
            # Without a task definition,
            # we can only update an existing service.
            self.log.error(
                'Could not find service with name {} to update '
                'in {}.'.format(
                    service_name, self._format_location()))
            raise exceptions.RecoverableActorFailure(
                'Cannot update non-existant service.')

        if service_is_active:
            # Service already exists - update it.
            # Ensure no immutable fields were mutated.
            self._check_immutable_field_errors(
                old_params=existing_service,
                new_params=self.service_definition,
                immutable_fields=self._immutable_fields())
            override = None
            if not self.option('use_existing_count'):
                override = {'desiredCount': self.option('count')}
            task_definition_name = yield self._update_service(
                service_name, existing_service, override=override)
        else:
            task_definition_name = yield self._create_service(service_name)

        if self.option('wait'):
            yield self._wait_for_service_update(
                service_name, task_definition_name)
            self.log.info(
                'Service {} updated successfully in {}.'.format(
                    service_name, self._format_location()))
        else:
            self.log.info(
                'Not waiting for service {} to finish updating '
                'in {}.'.format(
                    service_name, self._format_location()))

    @gen.coroutine
    def _execute(self):
        # Get service name
        if self.task_definition:
            service_name = (
                self.option('service_name') or
                self.task_definition['family'])
        else:
            service_name = self.option('service_name')

        if not service_name:
            raise exceptions.RecoverableActorFailure(
                'Either service_name or task_definition must be provided.')
        # Get existing service
        try:
            existing_service = yield self._describe_service(service_name)
        except ServiceNotFound:
            existing_service = None

        # Ensure state
        desired_state = self.option('state')
        if desired_state == 'present':
            yield self._ensure_service_present(service_name, existing_service)
        elif desired_state == 'absent':
            yield self._ensure_service_absent(service_name, existing_service)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.iam.certs`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging

from boto.exception import BotoServerError
from tornado import concurrent
from tornado import gen

from kingpin.actors import exceptions
from kingpin.actors.aws.iam import base
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class UploadCert(base.IAMBaseActor):

    """Uploads a new SSL Cert to AWS IAM.

    **Options**

    :private_key_path:
      (str) Path to the private key.

    :path:
      (str) The AWS "path" for the server certificate. Default: "/"

    :public_key_path:
      (str) Path to the public key certificate.

    :name:
      (str) The name for the server certificate.

    :cert_chain_path:
      (str) Path to the certificate chain. Optional.

    **Example**

    .. code-block:: json

       { "actor": "aws.iam.UploadCert",
         "desc": "Upload a new cert",
         "options": {
           "name": "new-cert",
           "private_key_path": "/cert.key",
           "public_key_path": "/cert.pem",
           "cert_chain_path": "/cert-chain.pem"
         }
       }

    **Dry run**

    Checks that the passed file paths are valid. In the future will also
    validate that the files are of correct format and content.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name for the server certificate.'),
        'public_key_path': (str, REQUIRED,
                            'Path to the public key certificate.'),
        'private_key_path': (str, REQUIRED, 'Path to the private key.'),
        'cert_chain_path': (str, None, 'Path to the certificate chain.'),
        'path': (str, None, 'The path for the server certificate.')
    }

    @gen.coroutine
    def _upload(self, cert_name, cert_body, private_key, cert_chain, path):
        """Create a new server certificate in AWS IAM."""
        yield self.api_call(
            self.iam_conn.upload_server_cert,
            cert_name=cert_name,
            cert_body=cert_body,
            private_key=private_key,
            cert_chain=cert_chain,
            path=path)

    @gen.coroutine
    def _execute(self):
        """Gather all the cert data and upload it.

        The `boto` library requires actual cert contents, but this actor
        expects paths to files.
        """
        # Gather needed cert data
        cert_chain_body = None
        if self.option('cert_chain_path'):
            cert_chain_body = self.readfile(self.option('cert_chain_path'))

        cert_body = self.readfile(self.option('public_key_path'))
        private_key = self.readfile(self.option('private_key_path'))

        # Upload it
        if self._dry:
            self.log.info('Would upload cert "%s"' % self.option('name'))
            raise gen.Return()

        self.log.info('Uploading cert "%s"' % self.option('name'))
        yield self._upload(
            cert_name=self.option('name'),
            cert_body=cert_body,
            private_key=private_key,
            cert_chain=cert_chain_body,
            path=self.option('path'))


class DeleteCert(base.IAMBaseActor):

    """Delete an existing SSL Cert in AWS IAM.

    **Options**

    :name:
      (str) The name for the server certificate.

    **Example**

    .. code-block:: json

       { "actor": "aws.iam.DeleteCert",
         "desc": "Run DeleteCert",
         "options": {
           "name": "fill-in"
         }
       }

    **Dry run**

    Will find the cert by name or raise an exception if it's not found.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name for the server certificate.')
    }

    @gen.coroutine
    def _find_cert(self, name):
        """Find a cert by name."""

        self.log.debug('Searching for cert "%s"...' % name)
        try:
            yield self.api_call(self.iam_conn.get_server_certificate, name)
        except BotoServerError as e:
            raise exceptions.UnrecoverableActorFailure(
                'Could not find cert %s. Reason: %s' % (name, e))

    @gen.coroutine
    def _delete(self, cert_name):
        """Delete a server certificate in AWS IAM."""
        yield self.api_call(self.iam_conn.delete_server_cert, cert_name)

    @gen.coroutine
    def _execute(self):
        if self._dry:
            self.log.info('Checking that the cert exists...')
            yield self._find_cert(self.option('name'))
            self.log.info('Would delete cert "%s"' % self.option('name'))
            raise gen.Return()

        self.log.info('Deleting cert "%s"' % self.option('name'))
        yield self._delete(cert_name=self.option('name'))



import datetime
import logging
import mock

from boto3.exceptions import Boto3Error
from tornado import gen
from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors.aws import ecs as ecs_actor
from kingpin.actors.aws import settings
from kingpin.actors.test import helper
import importlib

log = logging.getLogger(__name__)


class TestECSBaseActor(testing.AsyncTestCase):
    def setUp(self):
        super(TestECSBaseActor, self).setUp()
        importlib.reload(ecs_actor)

    @testing.gen_test
    def test_int_count(self):
        self.actor = ecs_actor.ECSBaseActor(options={'count': 1})
        self.assertEqual(self.actor.option('count'), 1)
        self.assertEqual(type(self.actor.option('count')), int)

    @testing.gen_test
    def test_str_count(self):
        self.actor = ecs_actor.ECSBaseActor(options={'count': str('1')})
        self.assertEqual(self.actor.option('count'), 1)
        self.assertEqual(type(self.actor.option('count')), int)

    @testing.gen_test
    def test_unicode_count(self):
        self.actor = ecs_actor.ECSBaseActor(options={'count': str('1')})
        self.assertEqual(self.actor.option('count'), 1)
        self.assertEqual(type(self.actor.option('count')), int)

    @testing.gen_test
    def test_invalid_count(self):
        with self.assertRaises(exceptions.RecoverableActorFailure):
            ecs_actor.ECSBaseActor(options={'count': str('?')})


class TestHandleFailures(testing.AsyncTestCase):

    def setUp(self):
        super(TestHandleFailures, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = ecs_actor.ECSBaseActor()

    @testing.gen_test
    def test_no_failures(self):
        failures = []
        self.actor._handle_failures(failures)

    @testing.gen_test
    def test_ignored_failure(self):
        ignore = 'ignore'
        failures = [{'reason': ignore}]
        self.actor._handle_failures(failures, ignore)

    @testing.gen_test
    def test_one_failure(self):
        failures = [{'reason': 1}]
        with self.assertRaises(exceptions.RecoverableActorFailure):
            self.actor._handle_failures(failures)

    @testing.gen_test
    def test_multiple_failures(self):
        failures = [{'reason': 1}, {'reason': 2}]
        with self.assertRaises(exceptions.RecoverableActorFailure):
            self.actor._handle_failures(failures)


class TestLoadTaskDefinition(testing.AsyncTestCase):

    def setUp(self):
        super(TestLoadTaskDefinition, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = ecs_actor.ECSBaseActor()

    @testing.gen_test
    def test_schema_ok_minimal(self):
        self._test({
            'family': '',
            'containerDefinitions': []})

    @testing.gen_test
    def test_schema_ok_complex(self):
        self._test({
            'family': 'family-name',
            'containerDefinitions': [
                {
                    'name': 'test',
                    'image': 'test-image',
                    'cpu': 1,
                    'memory': 2,
                    'links': ['a'],
                    'portMappings': [{
                        'containerPort': 8000,
                        'hostPort': 80,
                        'protocol': 'tcp'
                    }],
                    'essential': True,
                    'environment': [
                        {'name': 'var1', 'value': 'value1'},
                        {'name': 'var2', 'value': 'value2'}
                    ],
                },
                {
                    'name': 'a',
                    'image': 'a-image'
                }
            ],
            'volumes': [{
                'name': 'volume',
                'host': {
                    'sourcePath': '/mountme'
                }
            }]
        })

    @testing.gen_test
    def test_schema_ok_extra(self):
        self._test({
            'other stuff': 0,
            'family': 'name',
            'containerDefinitions': [],
            'extra': 'read all about it'})

    @testing.gen_test
    def test_schema_fail_empty(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self._test('')

    @testing.gen_test
    def test_schema_fail_none(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self._test(None)

    @testing.gen_test
    def test_none_task_definition(self):
        self.assertEqual(self.actor._load_task_definition(None, {}), None)

    def _test(self, task_definition):
        with mock.patch('kingpin.utils.convert_script_to_dict',
                        return_value=task_definition) as utils_convert_mock:
            task_definition_file = 'file'

            init_tokens = {
                'env_token': 'env_value',
                'default_token': 'default_value'}
            tokens = {
                'token': 'value',
                'default_token': 'override_value'}
            final_tokens = {
                'token': 'value',
                'default_token': 'override_value',
                'env_token': 'env_value'}

            self.actor._load_task_definition(
                task_definition_file, tokens, init_tokens)
            utils_convert_mock.assert_called_with(
                task_definition_file, final_tokens)


class TestDeregisterTaskDefinition(testing.AsyncTestCase):

    def setUp(self):
        super(TestDeregisterTaskDefinition, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = ecs_actor.ECSBaseActor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_call(self):
        task_definition_name = 'name'
        yield self.actor._deregister_task_definition(task_definition_name)
        self.assertEqual(
            self.actor.ecs_conn.deregister_task_definition.call_count, 1)


class TestDescribeTaskDefinition(testing.AsyncTestCase):

    def setUp(self):
        super(TestDescribeTaskDefinition, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = ecs_actor.ECSBaseActor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_call(self):
        task_definition = {'family': 'family'}
        task_tags = [
            {
                'key': 'foo',
                'value': 'bar'
            }
        ]
        task_definition_name = 'task_definition'
        self.actor.ecs_conn.describe_task_definition.return_value = {
            'taskDefinition': task_definition,
            'tags': task_tags,
            'ResponseMetadata': 'metadata'}
        result = yield self.actor._describe_task_definition(
            task_definition_name)
        self.assertEqual(result, {
            'taskDefinition': task_definition,
            'tags': task_tags})

        call_args = self.actor.ecs_conn.describe_task_definition.call_args
        expected = ({'taskDefinition': task_definition_name,
                     'include': ['TAGS'], },)
        self.assertEqual(call_args, expected)


class TestListTaskDefinitions(testing.AsyncTestCase):

    def setUp(self):
        super(TestListTaskDefinitions, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = ecs_actor.ECSBaseActor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_none(self):
        arns = []

        class Paginator(object):
            def paginate(self, *args, **kwargs):
                return [{'taskDefinitionArns': arns}]

        paginator = Paginator()
        self.actor.ecs_conn.get_paginator.return_value = paginator
        result = yield self.actor._list_task_definitions(
            status='ACTIVE', family_prefix='family')
        self.assertEqual(result, arns)

    @testing.gen_test
    def test_one_page(self):
        arns = [1, 2, 3]

        class Paginator(object):
            def paginate(self, *args, **kwargs):
                return [{'taskDefinitionArns': arns}]

        paginator = Paginator()
        self.actor.ecs_conn.get_paginator.return_value = paginator
        result = yield self.actor._list_task_definitions()
        self.assertEqual(result, arns)

    @testing.gen_test
    def test_multi_page(self):
        arn_pages = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

        class Paginator(object):
            def paginate(self, *args, **kwargs):
                return [{'taskDefinitionArns': arns} for arns in arn_pages]

        paginator = Paginator()
        self.actor.ecs_conn.get_paginator.return_value = paginator

        result = yield self.actor._list_task_definitions()
        self.assertEqual(result, sum(arn_pages, []))


class TestLoadServiceDefinition(testing.AsyncTestCase):

    def setUp(self):
        super(TestLoadServiceDefinition, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = ecs_actor.ECSBaseActor()

    @testing.gen_test
    def test_schema_ok_minimal(self):
        self._test({})

    @testing.gen_test
    def test_schema_ok_complex(self):
        self._test({
            'loadBalancers': [
                {
                    'loadBalancerName': 'elb',
                    'containerPort': 8000
                }
            ],
            'deploymentConfiguration': {
                'maximumPercent': 100,
                'minimumHealthyPercent': 2
            }
        })

    @testing.gen_test
    def test_schema_ok_extra(self):
        self._test({
            'other stuff': 0,
            'extra': 'read all about it'})

    @testing.gen_test
    def test_schema_fail_empty(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self._test('')

    @testing.gen_test
    def test_schema_fail_none(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self._test(None)

    @testing.gen_test
    def test_default(self):
        result = self.actor._load_service_definition(None, {})
        self.assertEqual(result, {})

        result = self.actor._load_service_definition('', {})
        self.assertEqual(result, {})

    def _test(self, service_definition):
        with mock.patch('kingpin.utils.convert_script_to_dict',
                        return_value=service_definition) as utils_convert_mock:
            service_definition_file = 'file'

            init_tokens = {
                'env_token': 'env_value',
                'default_token': 'default_value'}
            tokens = {
                'token': 'value',
                'default_token': 'override_value'}
            final_tokens = {
                'token': 'value',
                'default_token': 'override_value',
                'env_token': 'env_value'}

            self.actor._load_service_definition(service_definition_file,
                                                tokens, init_tokens)
            utils_convert_mock.assert_called_with(
                service_definition_file, final_tokens)


class TestRegisterTask(testing.AsyncTestCase):

    def setUp(self):
        super(TestRegisterTask, self).setUp()
        importlib.reload(ecs_actor)

        self.actor = _mock_task_actor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_ok_minimal(self):
        task_definition = {
            'family': '',
            'containerDefinitions': []}
        response_task_definition = task_definition.copy()
        response_task_definition.update({'revision': 100})
        self.actor.ecs_conn.register_task_definition.return_value = {
            'taskDefinition': response_task_definition
        }
        task_definition_name = yield self.actor._register_task(task_definition)
        self.actor.ecs_conn.register_task_definition.assert_called_with(
            **task_definition)
        self.assertEqual(task_definition_name,
                         '{}:{}'.format(task_definition['family'], '100'))

    @testing.gen_test
    def test_dry(self):
        self.actor._dry = True
        task_definition = {
            'family': 'name',
            'containerDefinitions': []}
        task_definition_name = yield self.actor._register_task(task_definition)
        self.assertEqual(task_definition_name, None)

    @testing.gen_test
    def test_internal_exception(self):
        self.actor.ecs_conn.register_task_definition.side_effect = Boto3Error
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._register_task({
                'family': 'name',
                'containerDefinitions': []})


class TestTaskRun(testing.AsyncTestCase):

    def setUp(self):
        super(TestTaskRun, self).setUp()
        self.actor = _mock_task_actor()
        self.actor.ecs_conn = mock.Mock()
        self.actor._handle_failures = mock.Mock()
        gen.sleep = helper.mock_tornado()

    def tearDown(self):
        importlib.reload(gen)

    def test_ok_minimal(self):
        self._test(
            family='family',
            revision=1,
            return_value={
                'failures': [],
                'tasks': [
                    {'taskArn': '1'}
                ]
            })

    def test_ok_complex(self):
        self.actor._options['cluster'] = 'cluster'
        self.actor._options['count'] = 5
        self._test(
            family='family',
            revision=100,
            return_value={
                'failures': [],
                'tasks': [
                    {'taskArn': '1'},
                    {'taskArn': '2'},
                    {'taskArn': '3'},
                ]
            })

    @testing.gen_test
    def _test(self, family, revision, return_value):
        def fail_twice(*args, **kwargs):
            fail_twice.call_count += 1
            if fail_twice.call_count > 2:
                return {'tasks': return_value['tasks'], 'failures': []}
            return {'failures': ['failure']}

        fail_twice.call_count = 0

        self.actor.ecs_conn.run_task = fail_twice

        tasks = yield self.actor._run_task('{}:{}'.format(family, revision))

        self.assertEqual(fail_twice.call_count, 3)

        expected_tasks = [t['taskArn'] for t in return_value['tasks']]
        self.assertEqual(tasks, expected_tasks)

    @testing.gen_test
    def test_with_failures(self):
        failures = ['failure one', 'failure two']
        self.actor.ecs_conn.run_task.return_value = {
            'failures': failures}
        exception = exceptions.RecoverableActorFailure()
        self.actor._handle_failures.side_effect = exception
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._run_task('')

    @testing.gen_test
    def test_dry(self):
        self.actor._dry = True
        yield self.actor._run_task('')
        self.assertEqual(self.actor.ecs_conn.run_task.call_count, 0)


class TestTaskWait(testing.AsyncTestCase):

    def setUp(self):
        super(TestTaskWait, self).setUp()

        self.actor = _mock_task_actor()
        gen.sleep = helper.mock_tornado()

    def tearDown(self):
        importlib.reload(gen)

    @testing.gen_test
    def test_return_for_empty(self):
        self.actor._tasks_done = helper.mock_tornado(True)
        yield self.actor._wait_for_tasks([])
        self.assertEqual(self.actor._tasks_done._call_count, 0)
        self.assertEqual(gen.sleep._call_count, 0)

    @testing.gen_test
    def test_return_for_none(self):
        self.actor._tasks_done = helper.mock_tornado(True)
        yield self.actor._wait_for_tasks(None)
        self.assertEqual(self.actor._tasks_done._call_count, 0)
        self.assertEqual(gen.sleep._call_count, 0)

    @testing.gen_test
    def test_instant_success(self):
        self.actor._tasks_done = helper.mock_tornado(True)
        yield self.actor._wait_for_tasks(['0'])
        self.assertEqual(self.actor._tasks_done._call_count, 1)
        self.assertEqual(gen.sleep._call_count, 0)

    @testing.gen_test
    def test_two_failures_before_success(self):
        @gen.coroutine
        def fail_twice(*args):
            fail_twice.call_count += 1
            return fail_twice.call_count > 2

        fail_twice.call_count = 0

        self.actor._tasks_done = fail_twice
        yield self.actor._wait_for_tasks(['0'])
        self.assertEqual(fail_twice.call_count, 3)
        self.assertEqual(gen.sleep._call_count, 2)


class TestTaskDone(testing.AsyncTestCase):

    def setUp(self):
        super(TestTaskDone, self).setUp()
        importlib.reload(ecs_actor)

        self.actor = _mock_task_actor()
        self.actor.ecs_conn = mock.Mock()
        self.actor.ecs_conn.describe_tasks.return_value = {
            'failures': [],
            'tasks': []}
        self.actor._handle_failures = mock.Mock()
        self.actor._get_containers_from_tasks = mock.Mock()

    def tearDown(self):
        importlib.reload(ecs_actor)

    @testing.gen_test
    def test_one_stopped_container(self):
        tasks = ['1']
        self.actor._get_containers_from_tasks.return_value = [{
            'taskArn': '1',
            'lastStatus': 'STOPPED',
            'exitCode': 0
        }]
        result = yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)
        self.assertTrue(result)

    @testing.gen_test
    def test_two_stopped_container(self):
        tasks = ['1', '2']
        self.actor._get_containers_from_tasks.return_value = [
            {
                'taskArn': '1',
                'lastStatus': 'STOPPED',
                'exitCode': 0
            },
            {
                'taskArn': '2',
                'lastStatus': 'STOPPED',
                'exitCode': 0
            }]
        result = yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)
        self.assertTrue(result)

    @testing.gen_test
    def test_one_failed_container(self):
        tasks = ['1']
        self.actor._get_containers_from_tasks.return_value = [{
            'taskArn': '1',
            'lastStatus': 'STOPPED',
            'exitCode': 100
        }]
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)

    @testing.gen_test
    def test_still_running(self):
        tasks = ['1']
        self.actor._get_containers_from_tasks.return_value = [{
            'taskArn': '1',
            'lastStatus': 'RUNNING'
        }]
        result = yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)
        self.assertFalse(result)

    @testing.gen_test
    def test_some_still_running(self):
        tasks = ['1']
        self.actor._get_containers_from_tasks.return_value = [
            {
                'taskArn': '1',
                'lastStatus': 'RUNNING'
            },
            {
                'taskArn': '2',
                'lastStatus': 'STOPPED',
                'exitCode': 0
            }]
        result = yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)
        self.assertFalse(result)

    @testing.gen_test
    def test_some_pass_some_fail(self):
        tasks = ['1', '2']
        self.actor._get_containers_from_tasks.return_value = [
            {
                'taskArn': '1',
                'lastStatus': 'STOPPED',
                'exitCode': 0
            },
            {
                'taskArn': '2',
                'lastStatus': 'STOPPED',
                'exitCode': 1
            }]
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)

    @testing.gen_test
    def test_fail_without_exit(self):
        tasks = ['1']
        self.actor._get_containers_from_tasks.return_value = [{
            'taskArn': '1',
            'lastStatus': 'STOPPED',
            'reason': 'fail reason'}]
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)

    @testing.gen_test
    def test_some_pass_some_fail_reverse(self):
        tasks = ['1', '2']
        self.actor._get_containers_from_tasks.return_value = [
            {
                'taskArn': '1',
                'lastStatus': 'STOPPED',
                'exitCode': 1
            },
            {
                'taskArn': '2',
                'lastStatus': 'STOPPED',
                'exitCode': 0
            }]
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._tasks_done(tasks)
        self.actor.ecs_conn.describe_tasks.assert_called_with(
            cluster=self.actor.option('cluster'),
            tasks=tasks)

    @testing.gen_test
    def test_failure_in_describe(self):
        tasks = ['1']
        failures = ['failure one', 'failure two']
        self.actor.ecs_conn.describe_tasks.return_value = {
            'failures': failures,
            'tasks': []
        }
        exception = exceptions.RecoverableActorFailure()
        self.actor._handle_failures.side_effect = exception
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._tasks_done(tasks)
        self.actor._handle_failures.assert_called_with(failures)


class TestTaskExecute(testing.AsyncTestCase):

    def setUp(self):
        super(TestTaskExecute, self).setUp()

        self.actor = _mock_task_actor()
        self.actor._register_task = helper.mock_tornado('family:1')
        self.actor._run_task = helper.mock_tornado([1, 2])
        self.actor._wait_for_tasks = helper.mock_tornado()

    @testing.gen_test
    def test_ok(self):
        yield self.actor._execute()
        self.assertEqual(self.actor._register_task._call_count, 1)
        self.assertEqual(self.actor._run_task._call_count, 1)
        self.assertEqual(self.actor._wait_for_tasks._call_count, 1)

    @testing.gen_test
    def test_without_wait(self):
        self.actor._options['wait'] = False
        yield self.actor._execute()
        self.assertEqual(self.actor._register_task._call_count, 1)
        self.assertEqual(self.actor._run_task._call_count, 1)
        self.assertEqual(self.actor._wait_for_tasks._call_count, 0)

    @testing.gen_test
    def test_dry(self):
        self.actor._dry = True
        yield self.actor._execute()
        self.assertEqual(self.actor._register_task._call_count, 1)
        self.assertEqual(self.actor._run_task._call_count, 1)
        self.assertEqual(self.actor._wait_for_tasks._call_count, 1)


class TestDescribeService(testing.AsyncTestCase):

    def setUp(self):
        super(TestDescribeService, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()
        self.actor._handle_failures = mock.Mock()

    @testing.gen_test
    def test_describe_call(self):
        self.actor.ecs_conn.describe_services.return_value = {
            'failures': [],
            'services': []}

        service_name1 = 'service_name1'
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._describe_service(service_name1)
        self.actor.ecs_conn.describe_services.assert_called_with(
            cluster=self.actor.option('cluster'),
            services=[service_name1])

        service_name2 = 'service_name2'
        self.actor._options['cluster'] = 'cluster'
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._describe_service(service_name2)
        self.actor.ecs_conn.describe_services.assert_called_with(
            cluster=self.actor.option('cluster'),
            services=[service_name2])

    @testing.gen_test
    def test_failure_call(self):
        service_name = 'service_name'

        failures = []
        self.actor.ecs_conn.describe_services.return_value = {
            'failures': failures,
            'services': []}

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._describe_service(service_name)
        self.actor._handle_failures.assert_called_with(
            failures,
            ecs_actor.ECSBaseActor.FAILURE_MISSING)

        failures = ['1', '2']
        self.actor.ecs_conn.describe_services.return_value = {
            'failures': failures,
            'services': []}

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._describe_service(service_name)
        self.actor._handle_failures.assert_called_with(
            failures,
            ecs_actor.ECSBaseActor.FAILURE_MISSING)

    @testing.gen_test
    def test_no_service(self):
        self.actor.ecs_conn.describe_services.return_value = {
            'failures': [],
            'services': []}

        service_name = 'service_name'
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._describe_service(service_name)

    @testing.gen_test
    def test_one_service(self):
        service_name = 'service_name'
        self.actor.ecs_conn.describe_services.return_value = {
            'failures': [],
            'services': [service_name]}

        result = yield self.actor._describe_service(service_name)
        self.assertEqual(result, service_name)

    @testing.gen_test
    def test_multiple_services_only_one(self):
        service_name1 = 'service_name1'
        service_name2 = 'service_name2'
        self.actor.ecs_conn.describe_services.return_value = {
            'failures': [],
            'services': [service_name1, service_name2]}

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._describe_service(service_name1)


class TestWaitForDeploymentUpdate(testing.AsyncTestCase):

    def setUp(self):
        super(TestWaitForDeploymentUpdate, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()
        gen.sleep = helper.mock_tornado()

    def tearDown(self):
        importlib.reload(gen)

    @testing.gen_test
    def test_slow_update(self):
        @gen.coroutine
        def service_not_found_once(*args, **kwargs):
            service_not_found_once.call_count += 1
            if service_not_found_once.call_count == 1:
                raise ecs_actor.ServiceNotFound()
            raise gen.Return({'status': 'ACTIVE'})

        service_not_found_once.call_count = 0

        def fail_twice(*args, **kwargs):
            fail_twice.call_count += 1
            if fail_twice.call_count > 2:
                return {'taskDefinition': 'arn/family:1'}

        fail_twice.call_count = 0

        self.actor._get_primary_deployment = fail_twice
        self.actor._describe_service = service_not_found_once

        yield self.actor._wait_for_deployment_update(
            service_name='service',
            task_definition_name='family:1')
        self.assertEqual(fail_twice.call_count, 3)


class TestIsTaskDefinitionDifferent(testing.AsyncTestCase):

    def setUp(self):
        super(TestIsTaskDefinitionDifferent, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_same(self):
        @gen.coroutine
        def mock_describe_task_definition(*args, **kwargs):
            mock_describe_task_definition.call_count += 1
            if mock_describe_task_definition.call_count == 1:
                raise gen.Return({'taskDefinition': {
                    'revision': 0,
                    'taskDefinitionArn': 'arn/family:1',
                    'family': 'family'
                }, 'tags': [{'key': 'foo', 'value': 'bar'}]
                })
            if mock_describe_task_definition.call_count == 2:
                raise gen.Return({'taskDefinition': {
                    'revision': 0,
                    'taskDefinitionArn': 'arn/family:2',
                    'family': 'family'
                }, 'tags': [{'key': 'foo', 'value': 'bar'}]
                })
            if mock_describe_task_definition.call_count > 2:
                self.fail('Called more than twice.')

        mock_describe_task_definition.call_count = 0
        self.actor._describe_task_definition = mock_describe_task_definition
        diff = yield self.actor._is_task_definition_different('a', 'b')
        self.assertEqual(diff, False)

    @testing.gen_test
    def test_different(self):
        @gen.coroutine
        def mock_describe_task_definition(*args, **kwargs):
            mock_describe_task_definition.call_count += 1
            if mock_describe_task_definition.call_count == 1:
                raise gen.Return({'taskDefinition': {
                    'revision': 0,
                    'taskDefinitionArn': 'arn/family:1',
                    'family': 'family1'
                }})
            if mock_describe_task_definition.call_count == 2:
                raise gen.Return({'taskDefinition': {
                    'revision': 0,
                    'taskDefinitionArn': 'arn/family:2',
                    'family': 'family2'
                }})
            if mock_describe_task_definition.call_count > 2:
                self.fail('Called more than twice.')

        mock_describe_task_definition.call_count = 0
        self.actor._describe_task_definition = mock_describe_task_definition
        diff = yield self.actor._is_task_definition_different('a', 'b')
        self.assertEquals(diff, True)

    @testing.gen_test
    def test_different_tags(self):
        @gen.coroutine
        def mock_describe_task_definition(*args, **kwargs):
            mock_describe_task_definition.call_count += 1
            if mock_describe_task_definition.call_count == 1:
                raise gen.Return({'taskDefinition': {
                    'revision': 0,
                    'taskDefinitionArn': 'arn/family:1',
                    'family': 'family'
                }, 'tags': [{'key': 'foo', 'value': 'bar'}]
                })
            if mock_describe_task_definition.call_count == 2:
                raise gen.Return({'taskDefinition': {
                    'revision': 0,
                    'taskDefinitionArn': 'arn/family:2',
                    'family': 'family'
                }, 'tags': [{'key': 'foo', 'value': 'baz'}]
                })
            if mock_describe_task_definition.call_count > 2:
                self.fail('Called more than twice.')

        mock_describe_task_definition.call_count = 0
        self.actor._describe_task_definition = mock_describe_task_definition
        diff = yield self.actor._is_task_definition_different('a', 'b')
        self.assertEqual(diff, True)


class TestCreateService(testing.AsyncTestCase):

    def setUp(self):
        super(TestCreateService, self).setUp()
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()
        self.actor._wait_for_deployment_update = helper.mock_tornado()

    @testing.gen_test
    def test_call(self):
        service_name = 'service_name'
        task_definition_name = 'family:1'
        self.actor._register_task = helper.mock_tornado(task_definition_name)

        yield self.actor._create_service(service_name=service_name)
        call_args = self.actor.ecs_conn.create_service.call_args
        expected = ({
                    'cluster': self.actor.option('cluster'),
                    'serviceName': service_name,
                    'taskDefinition': task_definition_name,
                    'desiredCount': self.actor.option('count'),
                    },)
        self.assertEqual(call_args, expected)

        task_definition_name = 'family:2'
        self.actor._register_task = helper.mock_tornado(task_definition_name)
        self.actor._options['cluster'] = 'cluster'
        self.actor._options['count'] = 5
        extra = {
            'test1': 'a',
            'test2': 'b'
        }
        self.actor.service_definition = extra
        yield self.actor._create_service(service_name=service_name)
        expected = ({
                    'cluster': self.actor.option('cluster'),
                    'serviceName': service_name,
                    'taskDefinition': task_definition_name,
                    'desiredCount': self.actor.option('count'),
                    },)
        expected[0].update(extra)
        call_args = self.actor.ecs_conn.create_service.call_args
        self.assertEqual(call_args, expected)

        self.assertEqual(self.actor._wait_for_deployment_update._call_count, 2)

    @testing.gen_test
    def test_dry(self):
        self.actor._dry = True
        yield self.actor._create_service(service_name='')
        self.assertFalse(self.actor.ecs_conn.create_service.called)

    @testing.gen_test
    def test_internal_exception(self):
        self.actor._register_task = helper.mock_tornado('family:1')
        self.actor.ecs_conn.create_service.side_effect = Boto3Error
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._create_service(service_name='service_name')


class TestStopService(testing.AsyncTestCase):
    def setUp(self):
        super(TestStopService, self).setUp()
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_call(self):
        self.actor._update_service = helper.mock_tornado()
        self.actor._wait_for_service_update = helper.mock_tornado()
        yield self.actor._stop_service('service',
                                       {'taskDefinition': 'arn/family:1'})

        self.assertEqual(self.actor._update_service._call_count, 1)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 1)


class TestDeleteService(testing.AsyncTestCase):
    def setUp(self):
        super(TestDeleteService, self).setUp()
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_active(self):
        self.actor._update_service = helper.mock_tornado()
        self.actor._wait_for_service_update = helper.mock_tornado()
        self.actor._stop_service = helper.mock_tornado()
        self.actor._list_task_definitions = helper.mock_tornado([1])
        self.actor._deregister_task_definition = helper.mock_tornado()
        yield self.actor._delete_service('service',
                                         {'status': 'ACTIVE'})
        self.assertEqual(self.actor.ecs_conn.delete_service.call_count, 1)
        self.assertEqual(self.actor._deregister_task_definition._call_count, 1)

    @testing.gen_test
    def test_deregister_false(self):
        self.actor._update_service = helper.mock_tornado()
        self.actor._wait_for_service_update = helper.mock_tornado()
        self.actor._stop_service = helper.mock_tornado()
        self.actor._list_task_definitions = helper.mock_tornado([1])
        self.actor._deregister_task_definition = helper.mock_tornado()
        yield self.actor._delete_service('service',
                                         {'status': 'ACTIVE'},
                                         deregister=False)
        self.assertEqual(self.actor.ecs_conn.delete_service.call_count, 1)
        self.assertEqual(self.actor._deregister_task_definition._call_count, 0)

    @testing.gen_test
    def test_inactive(self):
        self.actor._update_service = helper.mock_tornado()
        self.actor._wait_for_service_update = helper.mock_tornado()
        self.actor._stop_service = helper.mock_tornado()
        self.actor._list_task_definitions = helper.mock_tornado([1, 2])
        self.actor._deregister_task_definition = helper.mock_tornado()
        yield self.actor._delete_service('service',
                                         {'status': 'INACTIVE'})
        self.assertEqual(self.actor.ecs_conn.delete_service.call_count, 0)
        self.assertEqual(self.actor._deregister_task_definition._call_count, 2)

    @testing.gen_test
    def test_draining(self):
        self.actor._update_service = helper.mock_tornado()
        self.actor._wait_for_service_update = helper.mock_tornado()
        self.actor._stop_service = helper.mock_tornado()
        self.actor._list_task_definitions = helper.mock_tornado([1, 2])
        self.actor._deregister_task_definition = helper.mock_tornado()
        yield self.actor._delete_service('service',
                                         {'status': 'DRAINING'})
        self.assertEqual(self.actor.ecs_conn.delete_service.call_count, 0)
        self.assertEqual(self.actor._deregister_task_definition._call_count, 2)


class TestUpdateService(testing.AsyncTestCase):

    def setUp(self):
        super(TestUpdateService, self).setUp()
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()

    @testing.gen_test
    def test_call(self):
        service_name = 'service_name'
        configuration = 'configuration'
        existing_service = {'taskDefinition': 'arn/family:1',
                            'status': 'ACTIVE'}
        self.actor._register_task = helper.mock_tornado()
        self.actor._is_task_definition_different = helper.mock_tornado(False)
        self.actor.service_definition = {
            'deploymentConfiguration': configuration,
            'loadBalancers': 'loadBalancers',
            'role': 'role',
            'extra': 'included'}

        yield self.actor._update_service(
            service_name=service_name,
            existing_service=existing_service)
        call_args = self.actor.ecs_conn.update_service.call_args
        expected = ({
                    'cluster': self.actor.option('cluster'),
                    'service': service_name,
                    'deploymentConfiguration': configuration,
                    'extra': 'included'
                    },)
        self.assertEqual(call_args, expected)

        self.actor._options['cluster'] = 'cluster'
        yield self.actor._update_service(
            service_name=service_name,
            existing_service={'taskDefinition': 'arn/family:2',
                              'status': 'ACTIVE'})
        call_args = self.actor.ecs_conn.update_service.call_args
        expected = ({
                    'cluster': self.actor.option('cluster'),
                    'service': service_name,
                    'deploymentConfiguration': configuration,
                    'extra': 'included',
                    },)
        self.assertEqual(call_args, expected)
        self.assertEqual(self.actor._register_task._call_count, 2)
        self.assertEqual(self.actor._is_task_definition_different._call_count,
                         2)

    @testing.gen_test
    def test_different_task_definition(self):
        service_name = 'service_name'
        configuration = 'configuration'
        existing_service = {'taskDefinition': 'arn/family:1',
                            'status': 'ACTIVE'}
        self.actor._register_task = helper.mock_tornado('arn/family:1')
        self.actor._is_task_definition_different = helper.mock_tornado(True)
        self.actor._wait_for_deployment_update = helper.mock_tornado()
        self.actor.service_definition = {
            'deploymentConfiguration': configuration,
            'extra': 'included'}

        yield self.actor._update_service(
            service_name=service_name,
            existing_service=existing_service)
        call_args = self.actor.ecs_conn.update_service.call_args
        expected = ({
                    'cluster': self.actor.option('cluster'),
                    'service': service_name,
                    'taskDefinition': 'arn/family:1',
                    'deploymentConfiguration': configuration,
                    'extra': 'included'
                    },)
        self.assertEqual(call_args, expected)

        self.assertEqual(self.actor._wait_for_deployment_update._call_count, 1)

    @testing.gen_test
    def test_override(self):
        service_name = 'service_name'
        configuration = 'configuration'
        existing_service = {'taskDefinition': 'arn/family:1',
                            'status': 'ACTIVE'}
        self.actor._register_task = helper.mock_tornado()
        self.actor._is_task_definition_different = helper.mock_tornado(False)
        self.actor.service_definition = {
            'deploymentConfiguration': configuration,
            'extra': 'included'}

        yield self.actor._update_service(
            service_name=service_name,
            existing_service=existing_service,
            override={'desiredCount': 5})
        call_args = self.actor.ecs_conn.update_service.call_args
        expected = ({
                    'cluster': self.actor.option('cluster'),
                    'service': service_name,
                    'desiredCount': 5,
                    'deploymentConfiguration': configuration,
                    'extra': 'included'
                    },)
        self.assertEqual(call_args, expected)

    @testing.gen_test
    def test_dry(self):
        self.actor._dry = True
        yield self.actor._update_service(
            service_name='',
            existing_service=None)
        self.assertFalse(self.actor.ecs_conn.update_service.called)

    @testing.gen_test
    def test_internal_exception(self):
        existing_service = {'taskDefinition': 'arn/family:1',
                            'status': 'ACTIVE'}
        self.actor.ecs_conn.update_service.side_effect = Boto3Error
        self.actor._register_task = helper.mock_tornado()
        self.actor._is_task_definition_different = helper.mock_tornado(False)
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._update_service(
                service_name='service_name',
                existing_service=existing_service)

    @testing.gen_test
    def test_service_not_active(self):
        existing_service = {'taskDefinition': 'arn/family:1',
                            'status': 'INACTIVE'}
        self.actor._get_primary_deployment = mock.Mock()
        self.actor._describe_service = mock.Mock()
        self.actor._get_primary_deployment.return_value = existing_service
        self.actor._register_task = helper.mock_tornado()
        self.actor._describe_service = helper.mock_tornado()
        self.actor._is_task_definition_different = helper.mock_tornado(False)
        yield self.actor._update_service(
            service_name='service_name',
            existing_service=existing_service)


class TestEnsureServicePresent(testing.AsyncTestCase):

    def setUp(self):
        super(TestEnsureServicePresent, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()
        self.actor._create_service = helper.mock_tornado()
        self.actor._update_service = helper.mock_tornado()
        self.actor._check_immutable_field_errors = mock.Mock()
        self.actor._wait_for_service_update = helper.mock_tornado()
        self.actor._get_primary_deployment = mock.Mock()
        self.actor._get_primary_deployment.return_value = {
            'taskDefinition': 'arn/family:1'}

        self.service_name = 'service_name'
        self.task_definition_name = 'family:1'
        self.start = datetime.datetime.now()

    @testing.gen_test
    def test_create(self):
        yield self.actor._ensure_service_present(
            service_name=self.service_name,
            existing_service=None)
        self.assertEqual(self.actor._create_service._call_count, 1)
        self.assertEqual(self.actor._update_service._call_count, 0)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 1)

    @testing.gen_test
    def test_create_inactive(self):
        yield self.actor._ensure_service_present(
            service_name=self.service_name,
            existing_service={'status': 'INACTIVE'})
        self.assertEqual(self.actor._create_service._call_count, 1)
        self.assertEqual(self.actor._update_service._call_count, 0)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 1)

    @testing.gen_test
    def test_update(self):
        self.actor._check_immutable_field_errors.return_value = []
        yield self.actor._ensure_service_present(
            service_name=self.service_name,
            existing_service={'status': 'ACTIVE'})
        self.assertEqual(self.actor._create_service._call_count, 0)
        self.assertEqual(self.actor._update_service._call_count, 1)
        self.assertEqual(self.actor._check_immutable_field_errors.call_count,
                         1)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 1)

    @testing.gen_test
    def test_update_with_immutable_error(self):
        ex = exceptions.RecoverableActorFailure
        self.actor._check_immutable_field_errors.side_effect = ex
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._ensure_service_present(
                service_name=self.service_name,
                existing_service={'status': 'ACTIVE'})
        self.assertEqual(self.actor._create_service._call_count, 0)
        self.assertEqual(self.actor._update_service._call_count, 0)
        self.assertEqual(self.actor._check_immutable_field_errors.call_count,
                         1)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 0)

    @testing.gen_test
    def test_non_existant_update_error(self):
        self.actor.task_definition = None
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._ensure_service_present(
                service_name=self.service_name,
                existing_service=None)
        self.assertEqual(self.actor._create_service._call_count, 0)
        self.assertEqual(self.actor._update_service._call_count, 0)
        self.assertEqual(self.actor._check_immutable_field_errors.call_count,
                         0)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 0)

    @testing.gen_test
    def test_no_wait(self):
        self.actor._options['wait'] = False
        yield self.actor._ensure_service_present(
            service_name=self.service_name,
            existing_service={'status': 'ACTIVE'})
        self.assertEqual(self.actor._create_service._call_count, 0)
        self.assertEqual(self.actor._update_service._call_count, 1)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 0)

    @testing.gen_test
    def test_not_use_existing_count(self):
        self.actor._options['use_existing_count'] = False
        yield self.actor._ensure_service_present(
            service_name=self.service_name,
            existing_service={'status': 'ACTIVE'})
        self.assertEqual(self.actor._create_service._call_count, 0)
        self.assertEqual(self.actor._update_service._call_count, 1)
        self.assertEqual(self.actor._wait_for_service_update._call_count, 1)


class TestEnsureServiceAbsent(testing.AsyncTestCase):

    def setUp(self):
        super(TestEnsureServiceAbsent, self).setUp()
        importlib.reload(ecs_actor)
        self.actor = _mock_service_actor()
        self.actor.ecs_conn = mock.Mock()
        self.actor._delete_service = helper.mock_tornado()

    @testing.gen_test
    def test_existing(self):
        yield self.actor._ensure_service_absent(
            service_name='service',
            existing_service={'family': 'family'})
        self.assertEqual(self.actor._delete_service._call_count, 1)

    @testing.gen_test
    def test_absent(self):
        yield self.actor._ensure_service_absent(
            service_name='service',
            existing_service=None)
        self.assertEqual(self.actor._delete_service._call_count, 0)


class TestWaitForServiceUpdate(testing.AsyncTestCase):

    def setUp(self):
        super(TestWaitForServiceUpdate, self).setUp()

        self.actor = _mock_service_actor()
        gen.sleep = helper.mock_tornado()

        self.service_name = 'service_name'

    def tearDown(self):
        importlib.reload(gen)

    @testing.gen_test
    def test_instant_success(self):
        self.actor._is_service_updated = helper.mock_tornado(True)

        yield self.actor._wait_for_service_update(
            self.service_name, 'family:1')
        self.assertEqual(self.actor._is_service_updated._call_count, 1)
        self.assertEqual(gen.sleep._call_count, 0)

    @testing.gen_test
    def test_two_failures_before_success(self):
        @gen.coroutine
        def fail_twice(*args):
            fail_twice.call_count += 1
            return fail_twice.call_count > 2

        fail_twice.call_count = 0

        self.actor._is_service_updated = fail_twice
        yield self.actor._wait_for_service_update(
            self.service_name, 'family:1')
        self.assertEqual(fail_twice.call_count, 3)
        self.assertEqual(gen.sleep._call_count, 2)


class TestIsServiceUpdated(testing.AsyncTestCase):

    def setUp(self):
        super(TestIsServiceUpdated, self).setUp()

        self.actor = _mock_service_actor()
        self.actor._get_sorted_new_log_events = mock.Mock()
        self.actor._get_sorted_new_log_events.return_value = []

        self.service_name = 'service_name'
        self.start = datetime.datetime.now()

    @testing.gen_test
    def test_no_deployments(self):
        self.actor._describe_service = helper.mock_tornado({'deployments': []})
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._is_service_updated(
                service_name=self.service_name,
                task_definition_name='family:1')

        self.assertEqual(self.actor._describe_service._call_count, 1)

    @testing.gen_test
    def test_no_primary_deployment(self):
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'NOT PRIMARY'
            }]})
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._is_service_updated(
                service_name=self.service_name,
                task_definition_name='family:1')

        self.assertEqual(self.actor._describe_service._call_count, 1)

    @testing.gen_test
    def test_done(self):
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'PRIMARY',
                'taskDefinition': 'arn/family:1',
                'createdAt': self.start,
                'updatedAt': self.start,
                'runningCount': 1,
                'desiredCount': 1}],
            'events': []})
        result = yield self.actor._is_service_updated(
            service_name=self.service_name,
            task_definition_name='family:1')
        self.assertTrue(result)

    @testing.gen_test
    def test_not_done(self):
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'PRIMARY',
                'taskDefinition': 'arn/family:1',
                'createdAt': self.start,
                'updatedAt': self.start,
                'runningCount': 0,
                'desiredCount': 1}],
            'events': []})
        result = yield self.actor._is_service_updated(
            service_name=self.service_name,
            task_definition_name='family:1')
        self.assertFalse(result)

    @testing.gen_test
    def test_not_done_extra_deployments(self):
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'PRIMARY',
                'taskDefinition': 'arn/family:1',
                'createdAt': self.start,
                'updatedAt': self.start,
                'runningCount': 0,
                'desiredCount': 1,
            }, {
                'status': 'NOT PRIMARY',
            }],
            'events': []})
        result = yield self.actor._is_service_updated(
            service_name=self.service_name,
            task_definition_name='family:1')
        self.assertFalse(result)

    @testing.gen_test
    def test_done_extra_deployments(self):
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'PRIMARY',
                'taskDefinition': 'arn/family:1',
                'createdAt': self.start,
                'updatedAt': self.start,
                'runningCount': 1,
                'desiredCount': 1,
            }, {
                'status': 'NOT PRIMARY',
            }],
            'events': []})
        result = yield self.actor._is_service_updated(
            service_name=self.service_name,
            task_definition_name='family:1')
        self.assertFalse(result)

    @testing.gen_test
    def test_events(self):
        events = ['event1', 'event2']
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'PRIMARY',
                'taskDefinition': 'arn/family:1',
                'createdAt': self.start,
                'updatedAt': self.start,
                'runningCount': 1,
                'desiredCount': 1}],
            'events': events})
        self.actor._get_sorted_new_log_events.return_value = [
            (self.start, events[0]),
            (self.start, events[1])]
        result = yield self.actor._is_service_updated(
            service_name=self.service_name,
            task_definition_name='family:1')
        self.actor._get_sorted_new_log_events.assert_called_with(
            events=events,
            start_timestamp=self.start)
        self.assertTrue(result)

    @testing.gen_test
    def test_stale_response(self):
        self.actor._describe_service = helper.mock_tornado({
            'deployments': [{
                'status': 'PRIMARY',
                'taskDefinition': 'arn/family:1',
                'createdAt': self.start,
                'updatedAt': self.start,
                'runningCount': 1,
                'desiredCount': 1}],
            'events': []})
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._is_service_updated(
                service_name=self.service_name,
                task_definition_name='family:2')


class TestServiceExecute(testing.AsyncTestCase):

    def setUp(self):
        super(TestServiceExecute, self).setUp()

        self.actor = _mock_service_actor()
        self.actor._ensure_service_present = helper.mock_tornado()
        self.actor._ensure_service_absent = helper.mock_tornado()
        self.actor._describe_service = helper.mock_tornado()

    @testing.gen_test
    def test_ok(self):
        yield self.actor._execute()
        self.assertEqual(self.actor._describe_service._call_count, 1)
        self.assertEqual(self.actor._ensure_service_present._call_count, 1)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 0)

    @testing.gen_test
    def test_dry(self):
        self.actor._dry = True
        yield self.actor._execute()
        self.assertEqual(self.actor._describe_service._call_count, 1)
        self.assertEqual(self.actor._ensure_service_present._call_count, 1)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 0)

    @testing.gen_test
    def test_state_absent(self):
        self.actor._options['state'] = 'absent'
        self.actor._describe_service = helper.mock_tornado('service')
        yield self.actor._execute()
        self.assertEqual(self.actor._describe_service._call_count, 1)
        self.assertEqual(self.actor._ensure_service_present._call_count, 0)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 1)

    @testing.gen_test
    def test_service_already_absent(self):
        self.actor._options['state'] = 'absent'
        self.actor._describe_service = helper.mock_tornado()
        yield self.actor._execute()
        self.assertEqual(self.actor._describe_service._call_count, 1)
        self.assertEqual(self.actor._ensure_service_present._call_count, 0)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 1)

    @testing.gen_test
    def test_no_task_definition(self):
        self.actor.task_definition = None
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._execute()
        self.assertEqual(self.actor._describe_service._call_count, 0)
        self.assertEqual(self.actor._ensure_service_present._call_count, 0)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 0)

    @testing.gen_test
    def test_no_task_definition_has_service_name(self):
        self.actor._options['service_name'] = 'service'
        self.actor.task_definition = None
        yield self.actor._execute()
        self.assertEqual(self.actor._describe_service._call_count, 1)
        self.assertEqual(self.actor._ensure_service_present._call_count, 1)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 0)

    @testing.gen_test
    def test_service_not_found(self):
        @gen.coroutine
        def raise_service_not_found(*args, **kwargs):
            raise ecs_actor.ServiceNotFound()

        self.actor._describe_service = raise_service_not_found
        yield self.actor._execute()
        self.assertEqual(self.actor._ensure_service_present._call_count, 1)
        self.assertEqual(self.actor._ensure_service_absent._call_count, 0)


class TestGetContainersFromTasks(testing.AsyncTestCase):

    def setUp(self):
        super(TestGetContainersFromTasks, self).setUp()
        self.actor = _mock_task_actor()

    @testing.gen_test
    def test_empty(self):
        result = self.actor._get_containers_from_tasks([])
        self.assertEqual(result, [])

    @testing.gen_test
    def test_one(self):
        result = self.actor._get_containers_from_tasks([{
            'containers': [1]
        }])
        self.assertEqual(result, [1])

    @testing.gen_test
    def test_two_in_one_container(self):
        result = self.actor._get_containers_from_tasks([{
            'containers': [1, 2]
        }])
        self.assertEqual(result, [1, 2])

    @testing.gen_test
    def test_two_in_two_containers(self):
        result = self.actor._get_containers_from_tasks([
            {
                'containers': [1, 2]
            },
            {
                'containers': [3, 4]
            }
        ])
        self.assertEqual(result, [1, 2, 3, 4])

    @testing.gen_test
    def test_objects(self):
        result = self.actor._get_containers_from_tasks([
            {
                'containers': [{'a', 'b'}, {'a': 1}]
            },
            {
                'containers': [1, 2]
            }
        ])
        self.assertEqual(result, [{'a', 'b'}, {'a': 1}, 1, 2])


class TestCheckImmutableFieldErrors(testing.AsyncTestCase):

    def setUp(self):
        super(TestCheckImmutableFieldErrors, self).setUp()
        self.actor = _mock_service_actor()
        self.actor.log.error = mock.Mock()

    @testing.gen_test
    def test_everything_empty(self):
        self._test(old={}, new={}, immutable=[], expected_error_count=0)

    @testing.gen_test
    def test_same_no_immutable(self):
        self._test(old={'a': 1},
                   new={'a': 1},
                   immutable=[], expected_error_count=0)

    @testing.gen_test
    def test_not_set_but_different(self):
        self._test(old={'a': 1, 'b': [1, 2, 3]},
                   new={'a': None, 'b': []},
                   immutable=['a', 'b'], expected_error_count=0)

    @testing.gen_test
    def test_different_no_immutable(self):
        self._test(old={'a': 1},
                   new={'b': 1},
                   immutable=[], expected_error_count=0)

    @testing.gen_test
    def test_same_only_has_immutable(self):
        self._test(old={'a': 1},
                   new={'a': 1},
                   immutable=['a'], expected_error_count=0)

    @testing.gen_test
    def test_different_only_has_immutable(self):
        self._test(old={'a': 1},
                   new={'a': 2},
                   immutable=['a'], expected_error_count=1)
        self._test(old={'a': 2},
                   new={'a': 1},
                   immutable=['a'], expected_error_count=1)

    @testing.gen_test
    def test_different_only_has_immutable_multiple(self):
        self._test(old={'a': 2},
                   new={'a': 1},
                   immutable=['a', 'b'], expected_error_count=1)
        self._test(old={'a': 2},
                   new={'a': 1},
                   immutable=['b', 'a'], expected_error_count=1)

    @testing.gen_test
    def test_complex(self):
        self._test(old={'a': 1, 'b': 2, 'c': 3},
                   new={'a': {'a': {'b': 1}}, 'b': 'different', 'c': 3},
                   immutable=['b', 'c'], expected_error_count=1)

    @testing.gen_test
    def test_role_logic(self):
        self._test(old={'roleArn': 'a/b'},
                   new={'role': 'b'},
                   immutable=['role'], expected_error_count=0)
        self._test(old={'roleArn': 'a/c'},
                   new={'role': 'b'},
                   immutable=['role'], expected_error_count=1)
        self._test(old={},
                   new={'role': 'b'},
                   immutable=['role'], expected_error_count=1)

    def _test(self, old, new, immutable, expected_error_count):
        # Reset call count
        self.actor.log.error.call_count = 0
        if expected_error_count > 0:
            with self.assertRaises(exceptions.RecoverableActorFailure):
                self.actor._check_immutable_field_errors(old, new, immutable)
            self.assertEqual(self.actor.log.error.call_count,
                             expected_error_count)
        else:
            self.actor._check_immutable_field_errors(old, new, immutable)


class TestGetSortedNewLogEvents(testing.AsyncTestCase):

    def setUp(self):
        super(TestGetSortedNewLogEvents, self).setUp()
        self.actor = _mock_service_actor()
        self.start = datetime.datetime.now()
        self.two_seconds_ago = self.start - datetime.timedelta(seconds=2)
        self.one_seconds_ago = self.start - datetime.timedelta(seconds=1)
        self.one_seconds_later = self.start + datetime.timedelta(seconds=1)
        self.two_seconds_later = self.start + datetime.timedelta(seconds=2)

    def test_no_events(self):
        self.actor.seen_events = set()
        events = self.actor._get_sorted_new_log_events(
            events=[],
            start_timestamp=self.start)
        self.assertEqual(events, [])

    def test_seen_all_one_event(self):
        self.actor.seen_events = {0}
        events = self.actor._get_sorted_new_log_events(
            events=[{
                'id': 0,
                'message': 'message',
                'createdAt': self.one_seconds_ago}],
            start_timestamp=self.start)
        self.assertEqual(events, [])

    def test_seen_all_two_events(self):
        self.actor.seen_events = {0, 1}
        events = self.actor._get_sorted_new_log_events(
            events=[{
                'id': 0,
                'message': 'message1',
                'createdAt': self.one_seconds_ago
            }, {
                'id': 1,
                'message': 'message2',
                'createdAt': self.one_seconds_later
            }],
            start_timestamp=self.start)
        self.assertEqual(events, [])

    def test_timestamp_filtering(self):
                    # If the supplied fields are an empty list, treat them also
                    # as an null field and move on.
        events_after = [{
            'id': 0,
            'message': 'message1',
            'createdAt': self.two_seconds_later
        }, {
            'id': 1,
            'message': 'message2',
            'createdAt': self.one_seconds_later
        }]
        events_before = [{
            'id': 2,
            'message': 'message3',
            'createdAt': self.one_seconds_ago}]
        self.actor.seen_events = set()
        events = self.actor._get_sorted_new_log_events(
            events=events_after + events_before,
            start_timestamp=self.start)
        expected = [
            (self.one_seconds_later, 'message2'),
            (self.two_seconds_later, 'message1')]
        self.assertEqual(events, expected)

    def test_seen_added_to(self):
        input_events = [{
            'id': 0,
            'message': 'message1',
            'createdAt': self.one_seconds_ago
        }, {
            'id': 1,
            'message': 'message2',
            'createdAt': self.one_seconds_later
        }]
        self.actor.seen_events = {2}
        events = self.actor._get_sorted_new_log_events(
            events=input_events,
            start_timestamp=self.start)
        expected = [
            (input_events[1]['createdAt'], input_events[1]['message'])
        ]
        self.assertEqual(events, expected)
        self.assertEqual(self.actor.seen_events, {1, 2})


def _mock_task_actor():
    settings.ECS_RETRY_ATTEMPTS = 0
    importlib.reload(ecs_actor)
    base_actor = 'kingpin.actors.aws.ecs.ECSBaseActor'
    load_task_definition = base_actor + '._load_task_definition'
    with mock.patch(load_task_definition):
        return ecs_actor.RunTask(
            options={
                'region': '',
                'cluster': '',
                'task_definition': ''})


def _mock_service_actor():
    settings.ECS_RETRY_ATTEMPTS = 0
    importlib.reload(ecs_actor)
    base_actor = 'kingpin.actors.aws.ecs.ECSBaseActor'
    load_task_definition = base_actor + '._load_task_definition'
    load_service_definition = base_actor + '._load_service_definition'
    with mock.patch(load_task_definition), mock.patch(load_service_definition):
        return ecs_actor.Service(
            options={
                'region': '',
                'cluster': '',
                'task_definition': ''})



"""Tests for the actors.rollbar package"""

import json
import mock
import io

from tornado import gen
from tornado import httpclient
from tornado import testing

from kingpin.actors import rollbar
from kingpin.actors import exceptions


__author__ = 'Matt Wise <matt@nextdoor.com>'


class FakeHTTPClientClass(object):

    '''Fake HTTPClient object for testing'''

    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        raise gen.Return(self.response_value)


class FakeExceptionRaisingHTTPClientClass(object):

    '''Fake HTTPClient object for testing'''

    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        raise self.response_value


class TestRollbarBase(testing.AsyncTestCase):

    """Unit tests for the Rollbar base actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestRollbarBase, self).setUp()
        rollbar.TOKEN = 'Unittest'

    def test_build_potential_args(self):
        potential_args = {
            'foo': 'bar',
            'baz': 'bat',
        }
        expected_args = dict(potential_args)
        expected_args['access_token'] = rollbar.TOKEN

        actor = rollbar.RollbarBase('Unittest Deploy', {})
        args = actor._build_potential_args(potential_args)

        self.assertEqual(args, expected_args)

    @testing.gen_test
    def test_init_without_environment_creds(self):
        # Un-set the token now and make sure the init fails
        rollbar.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            rollbar.RollbarBase('Unittest Deploy', {})

    @testing.gen_test
    def test_fetch_wrapper(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})

        # Valid response test
        response_dict = {'status': 'sent'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=200,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._fetch_wrapper('http://fake.com')
            self.assertEqual(res, response_dict)

    @testing.gen_test
    def test_fetch_wrapper_with_401(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})
        response_dict = {'err': 1, 'messsage': 'Unauthorized'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=401, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.InvalidCredentials):
                yield actor._fetch_wrapper('http://fake.com')

    @testing.gen_test
    def test_fetch_wrapper_with_403(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})
        response_dict = {'err': 1, 'messsage': 'access token not found: xxx'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=403, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.InvalidCredentials):
                yield actor._fetch_wrapper('http://fake.com')

    @testing.gen_test
    def test_fetch_wrapper_with_422(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})
        response_dict = {'err': 1, 'messsage': 'Unprocessable Entity'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=422, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.RecoverableActorFailure):
                yield actor._fetch_wrapper('http://fake.com')

    @testing.gen_test
    def test_fetch_wrapper_with_429(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})
        response_dict = {'err': 1, 'messsage': 'Too Many Requests'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=429, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.RecoverableActorFailure):
                yield actor._fetch_wrapper('http://fake.com')

    @testing.gen_test
    def test_fetch_wrapper_with_other_failure(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})
        response_dict = {'err': 1, 'messsage': 'Something bad happened'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=123, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.RecoverableActorFailure):
                yield actor._fetch_wrapper('http://fake.com')

    @testing.gen_test
    def test_project(self):
        actor = rollbar.RollbarBase('Unit Test Action', {})
        response_dict = {'err': 0, 'result': '...'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=200,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._project()
        self.assertEqual(res, response_dict)


class TestDeploy(testing.AsyncTestCase):

    """Unit tests for the Rollbar Deploy actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestDeploy, self).setUp()
        rollbar.TOKEN = 'Unittest'

    @testing.gen_test
    def test_deploy(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'unittest',
             'revision': '0001a',
             'local_username': 'bob',
             'comment': 'weeee'})

        @gen.coroutine
        def fake_fetch_wrapper(*args, **kwargs):
            raise gen.Return('123')
        actor._fetch_wrapper = fake_fetch_wrapper
        res = yield actor._deploy()
        self.assertEqual(res, '123')

    @testing.gen_test
    def test_deploy_with_rollbar_username(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'unittest',
             'revision': '0001a',
             'local_username': 'bob',
             'rollbar_username': 'bob@rollbar.com',
             'comment': 'weeee'})

        @gen.coroutine
        def fake_fetch_wrapper(*args, **kwargs):
            raise gen.Return('123')
        actor._fetch_wrapper = fake_fetch_wrapper
        res = yield actor._deploy()
        self.assertEqual(res, '123')

    @testing.gen_test
    def test_execute_dry(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'unittest',
             'revision': '0001a',
             'local_username': 'bob',
             'rollbar_username': 'bob@rollbar.com',
             'comment': 'weeee'}, dry=True)

        @gen.coroutine
        def fake_project(*args, **kwargs):
            raise gen.Return('123')
        actor._project = fake_project
        res = yield actor._execute()
        self.assertEqual(res, None)

    @testing.gen_test
    def test_execute(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'unittest',
             'revision': '0001a',
             'local_username': 'bob',
             'rollbar_username': 'bob@rollbar.com',
             'comment': 'weeee'})

        @gen.coroutine
        def fake_deploy(*args, **kwargs):
            raise gen.Return('123')
        actor._deploy = fake_deploy
        res = yield actor._execute()
        self.assertEqual(res, None)
